#!/usr/bin/env python3
"""Classes and methods for working with nondeterministic pushdown automata."""

import automata.base.exceptions as exceptions
import automata.pda.pda as pda
from automata.pda.configuration import PDAConfiguration
from automata.pda.stack import PDAStack


class NPDA(pda.PDA):
    """A nondeterministic pushdown automaton."""

    __slots__ = ('states', 'input_symbols', 'stack_symbols',
                 'transitions', 'initial_state',
                 'initial_stack_symbol', 'final_states', 'acceptance_mode')

    def __init__(self, *, states, input_symbols, stack_symbols,
                 transitions, initial_state,
                 initial_stack_symbol, final_states, acceptance_mode='both'):
        """Initialize a complete NPDA."""
        super().__init__(
            states=states,
            input_symbols=input_symbols,
            stack_symbols=stack_symbols,
            transitions=transitions,
            initial_state=initial_state,
            initial_stack_symbol=initial_stack_symbol,
            final_states=final_states,
            acceptance_mode=acceptance_mode
        )

    def _validate_transition_invalid_symbols(self, start_state, paths):
        """Raise an error if transition symbols are invalid."""
        for input_symbol, symbol_paths in paths.items():
            self._validate_transition_invalid_input_symbols(
                start_state, input_symbol)
            for stack_symbol in symbol_paths:
                self._validate_transition_invalid_stack_symbols(
                    start_state, stack_symbol)

    def _get_transitions(self, state, input_symbol, stack_symbol):
        """Get the transition tuples for the given state and symbols."""
        transitions = set()
        if (state in self.transitions and
                input_symbol in self.transitions[state] and
                stack_symbol in self.transitions[state][input_symbol]):
            for (
                dest_state,
                new_stack_top
            ) in self.transitions[state][input_symbol][stack_symbol]:
                transitions.add((
                    input_symbol,
                    dest_state,
                    new_stack_top
                ))
        return transitions

    def _get_next_configurations(self, old_config):
        """Advance to the next configurations."""
        transitions = set()
        if old_config.remaining_input:
            transitions.update(self._get_transitions(
                old_config.state,
                old_config.remaining_input[0],
                old_config.stack.top()
            ))
        transitions.update(self._get_transitions(
            old_config.state,
            '',
            old_config.stack.top()
        ))
        new_configs = set()
        for input_symbol, new_state, new_stack_top in transitions:
            remaining_input = old_config.remaining_input
            if input_symbol:
                remaining_input = remaining_input[1:]
            new_config = PDAConfiguration(
                new_state,
                remaining_input,
                self._replace_stack_top(old_config.stack, new_stack_top)
            )
            new_configs.add(new_config)
        return new_configs

    def read_input_stepwise(self, input_str):
        """
        Check if the given string is accepted by this NPDA.

        Yield the NPDA's current configurations at each step.
        """
        current_configurations = set()
        current_configurations.add(PDAConfiguration(
            self.initial_state,
            input_str,
            PDAStack([self.initial_stack_symbol])
        ))

        yield current_configurations

        while current_configurations:
            new_configurations = set()
            for config in current_configurations:
                if self._has_accepted(config):
                    # One accepting configuration is enough.
                    return
                if config.remaining_input:
                    new_configurations.update(
                        self._get_next_configurations(config)
                    )
                elif self._has_lambda_transition(
                    config.state,
                    config.stack.top()
                ):
                    new_configurations.update(
                        self._get_next_configurations(config)
                    )
            current_configurations = new_configurations
            yield current_configurations

        raise exceptions.RejectionException(
            'the NPDA did not reach an accepting configuration'
        )



#!/usr/bin/env python3
"""Classes and methods for working with nondeterministic finite automata."""
from collections import deque
from itertools import chain, product

import networkx as nx
from frozendict import frozendict
from pydot import Dot, Edge, Node

import automata.base.exceptions as exceptions
import automata.fa.fa as fa
from automata.regex.parser import RESERVED_CHARACTERS, parse_regex


class NFA(fa.FA):
    """A nondeterministic finite automaton."""

    __slots__ = ('states', 'input_symbols', 'transitions',
                 'initial_state', 'final_states')

    def __init__(self, *, states, input_symbols, transitions,
                 initial_state, final_states):
        """Initialize a complete NFA."""
        super().__init__(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=initial_state,
            final_states=final_states,
            _lambda_closures=self._compute_lambda_closures(states, transitions)
        )

    def _compute_lambda_closures(self, states, transitions):
        """
        Computes a dictionary of the lambda closures for this NFA, where each
        key is the state name and the value is the lambda closure for that
        corresponding state. This dictionary is cached for the lifetime of the
        instance, and is available via the 'lambda_closures' attribute.

        The lambda closure of a state q is the set containing q, along with
        every state that can be reached from q by following only lambda
        transitions.
        """
        lambda_graph = nx.DiGraph()
        lambda_graph.add_nodes_from(states)
        lambda_graph.add_edges_from([
            (start_state, end_state)
            for start_state, transition in transitions.items()
            for char, end_states in transition.items()
            if char == ''
            for end_state in end_states
        ])

        return frozendict({
            state: frozenset(nx.descendants(lambda_graph, state) | {state})
            for state in states
        })

    def __add__(self, other):
        """Return the concatenation of this NFA and another NFA."""
        if isinstance(other, NFA):
            return self.concatenate(other)
        else:
            return NotImplemented

    def __or__(self, other):
        """Return the union of this NFA and another NFA."""
        if isinstance(other, NFA):
            return self.union(other)
        else:
            return NotImplemented

    def __and__(self, other):
        """Return the union of this NFA and another NFA."""
        if isinstance(other, NFA):
            return self.intersection(other)
        else:
            return NotImplemented

    def __reversed__(self):
        """Return the reversal of this DFA."""
        return self.reverse()

    @classmethod
    def from_dfa(cls, dfa):
        """Initialize this NFA as one equivalent to the given DFA."""
        nfa_transitions = {
            start_state: {
                input_symbol: {end_state}
                for input_symbol, end_state in paths.items()
            }
            for start_state, paths in dfa.transitions.items()
        }

        return cls(
            states=dfa.states, input_symbols=dfa.input_symbols,
            transitions=nfa_transitions, initial_state=dfa.initial_state,
            final_states=dfa.final_states)

    def _validate_transition_invalid_symbols(self, start_state, paths):
        """Raise an error if transition symbols are invalid."""
        for input_symbol in paths.keys():
            if input_symbol not in self.input_symbols and input_symbol != '':
                raise exceptions.InvalidSymbolError(
                    'state {} has invalid transition symbol {}'.format(
                        start_state, input_symbol))

    @classmethod
    def _from_symbol(cls, symbol, input_symbols=None):
        """Generate NFA from single symbol, `input symbols` may be passed to initialize input_symbols for NFA"""
        states = {0, 1}
        initial_state = 0
        if input_symbols is None:
            input_symbols = {symbol}
        transitions = {0: {symbol: {1}}}
        final_states = {1}

        return cls(
            states=states,
            input_symbols=input_symbols,
            initial_state=initial_state,
            transitions=transitions,
            final_states=final_states
        )

    @classmethod
    def from_regex(cls, regex, *, input_symbols=None):
        """Initialize this NFA as one equivalent to the given regular expression"""

        if input_symbols is None:
            input_symbols = set(regex) - RESERVED_CHARACTERS
        else:
            conflicting_symbols = RESERVED_CHARACTERS & input_symbols
            if conflicting_symbols:
                raise exceptions.InvalidSymbolError(f'Invalid input symbols: {conflicting_symbols}')

        nfa_builder = parse_regex(regex, input_symbols)

        return cls(
            states=frozenset(nfa_builder._transitions.keys()),
            input_symbols=input_symbols,
            transitions=nfa_builder._transitions,
            initial_state=nfa_builder._initial_state,
            final_states=nfa_builder._final_states
        )

    def _validate_transition_end_states(self, start_state, paths):
        """Raise an error if transition end states are invalid."""
        for end_states in paths.values():
            for end_state in end_states:
                if end_state not in self.states:
                    raise exceptions.InvalidStateError(
                        'end state {} for transition on {} is '
                        'not valid'.format(end_state, start_state))

    def validate(self):
        """Return True if this NFA is internally consistent."""
        for start_state, paths in self.transitions.items():
            self._validate_transition_invalid_symbols(start_state, paths)
            self._validate_transition_end_states(start_state, paths)
        self._validate_initial_state()
        self._validate_initial_state_transitions()
        self._validate_final_states()
        return True

    def _get_next_current_states(self, current_states, input_symbol):
        """Return the next set of current states given the current set."""
        next_current_states = set()

        for current_state in current_states:
            if current_state not in self.transitions:
                continue
            current_transition = self.transitions[current_state]
            for end_state in current_transition.get(input_symbol, {}):
                next_current_states.update(
                    self._lambda_closures[end_state])

        return frozenset(next_current_states)

    @staticmethod
    def _compute_reachable_states(initial_state, input_symbols, transitions):
        """Compute the states which are reachable from the initial state."""

        visited_set = set()
        queue = deque()

        queue.append(initial_state)
        visited_set.add(initial_state)

        while queue:
            state = queue.popleft()
            state_dict = transitions.get(state)

            if state_dict:
                for next_state in chain.from_iterable(dest for dest in state_dict.values()):
                    if next_state not in visited_set:
                        visited_set.add(next_state)
                        queue.append(next_state)

        return visited_set

    def _eliminate_lambda(self):
        """Internal helper function for eliminate lambda. Doesn't create final NFA."""

        # Create new transitions and final states for running this algorithm
        new_transitions = {
            state: {
                symbol: set(dest)
                for symbol, dest in paths.items()
            }
            for state, paths in self.transitions.items()
        }
        new_final_states = set(self.final_states)

        for state in self.states:
            lambda_enclosure = self._lambda_closures[state] - {state}
            for input_symbol in self.input_symbols:
                next_current_states = self._get_next_current_states(lambda_enclosure, input_symbol)

                # Don't do anything if no new current states
                if next_current_states:
                    state_transition_dict = new_transitions.setdefault(state, dict())

                    if input_symbol in state_transition_dict:
                        state_transition_dict[input_symbol].update(next_current_states)
                    else:
                        state_transition_dict[input_symbol] = next_current_states

            if (new_final_states & lambda_enclosure):
                new_final_states.add(state)

            if state in new_transitions:
                new_transitions[state].pop('', None)

        # Remove unreachable states
        reachable_states = NFA._compute_reachable_states(self.initial_state, self.input_symbols, new_transitions)
        reachable_final_states = reachable_states & new_final_states

        for state in self.states - reachable_states:
            new_transitions.pop(state, None)

        return reachable_states, new_transitions, reachable_final_states

    def eliminate_lambda(self):
        """Removes epsilon transitions from the NFA which recognizes the same language."""

        reachable_states, new_transitions, reachable_final_states = self._eliminate_lambda()

        return self.__class__(
            states=reachable_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=self.initial_state,
            final_states=reachable_final_states
        )

    def _check_for_input_rejection(self, current_states):
        """Raise an error if the given config indicates rejected input."""
        if not (current_states & self.final_states):
            raise exceptions.RejectionException(
                'the NFA stopped on all non-final states ({})'.format(
                    ', '.join(str(state) for state in current_states)))

    def read_input_stepwise(self, input_str):
        """
        Check if the given string is accepted by this NFA.

        Yield the current configuration of the NFA at each step.
        """
        current_states = self._lambda_closures[self.initial_state]

        yield current_states
        for input_symbol in input_str:
            current_states = self._get_next_current_states(
                current_states, input_symbol)
            yield current_states

        self._check_for_input_rejection(current_states)

    @staticmethod
    def _get_state_maps(state_set_a, state_set_b, *, start=0):
        """
        Generate state map dicts from given sets. Useful when the state set has
        to be a union of the state sets of component FAs.
        """

        state_map_a = {
            state: i
            for i, state in enumerate(state_set_a, start=start)
        }

        state_map_b = {
            state: i
            for i, state in enumerate(state_set_b, start=max(state_map_a.values())+1)
        }

        return (state_map_a, state_map_b)

    def union(self, other):
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the union of L1 and L2.
        """

        # Starting at 1 because 0 is for the initial state
        (state_map_a, state_map_b) = NFA._get_state_maps(self.states, other.states, start=1)

        new_states = frozenset(chain(state_map_a.values(), state_map_b.values(), [0]))
        new_transitions = {state: dict() for state in new_states}

        # Connect new initial state to both branch
        new_transitions[0] = {'': {state_map_a[self.initial_state], state_map_b[other.initial_state]}}

        # Transitions of self
        NFA._load_new_transition_dict(state_map_a, self.transitions, new_transitions)
        # Transitions of other
        NFA._load_new_transition_dict(state_map_b, other.transitions, new_transitions)

        # Final states
        new_final_states = frozenset(chain(
            (state_map_a[state] for state in self.final_states),
            (state_map_b[state] for state in other.final_states)
        ))

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols | other.input_symbols,
            transitions=new_transitions,
            initial_state=0,
            final_states=new_final_states
        )

    def concatenate(self, other):
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the languages L1 concatenated with L2.
        """

        (state_map_a, state_map_b) = NFA._get_state_maps(self.states, other.states)

        new_states = frozenset(chain(state_map_a.values(), state_map_b.values()))
        new_transitions = {state: dict() for state in new_states}

        # Transitions of self
        NFA._load_new_transition_dict(state_map_a, self.transitions, new_transitions)
        # Transitions of other
        NFA._load_new_transition_dict(state_map_b, other.transitions, new_transitions)

        # Transitions from self to other
        for state in self.final_states:
            new_transitions[state_map_a[state]].setdefault('', set()).add(
                state_map_b[other.initial_state]
            )

        # Final states of other
        new_final_states = frozenset(state_map_b[state] for state in other.final_states)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols | other.input_symbols,
            transitions=new_transitions,
            initial_state=state_map_a[self.initial_state],
            final_states=new_final_states
        )

    def kleene_star(self):
        """
        Given an NFA which accepts the language L returns
        an NFA which accepts L repeated 0 or more times.
        """
        new_states = set(self.states)
        new_initial_state = NFA._add_new_state(new_states)

        # Transitions are the same with a few additions.
        new_transitions = dict(self.transitions)
        # We add epsilon transition from new initial state
        # to old initial state
        new_transitions[new_initial_state] = {
            '': {self.initial_state}
        }
        # For each final state in original NFA we add epsilon
        # transition to the old initial state
        for state in self.final_states:
            new_transitions[state] = dict(new_transitions.get(state, dict()))
            transition = new_transitions[state]
            transition[''] = set(transition.get('', set()))
            transition[''].add(self.initial_state)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=self.final_states | {new_initial_state}
        )

    def option(self):
        """
        Given an NFA which accepts the language L returns
        an NFA which accepts L repeated 0 or 1 times. (option - ?)
        Note: still you cannot pass empty string to the machine.
        """
        new_states = set(self.states)
        new_initial_state = NFA._add_new_state(new_states)

        # Transitions are the same with a few additions.
        new_transitions = dict(self.transitions)
        # We add epsilon transition from new initial state
        # to old initial state
        new_transitions[new_initial_state] = {
            '': {self.initial_state}
        }

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=self.final_states | {new_initial_state}
        )

    def reverse(self):
        """
        Given an NFA which accepts the language L this function
        returns an NFA which accepts the reverse of L.
        """
        new_states = set(self.states)
        new_initial_state = NFA._add_new_state(new_states)

        # Transitions are the same except reversed
        new_transitions = {
            state: dict() for state in new_states
        }

        for state_a, transitions in self.transitions.items():
            for symbol, states in transitions.items():
                for state_b in states:
                    new_transitions[state_b].setdefault(symbol, set()).add(state_a)

        # And we additionally have epsilon transitions from
        # new initial state to each old final state.
        new_transitions[new_initial_state][''] = set(self.final_states)
        new_final_states = {self.initial_state}

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def intersection(self, other):
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the intersection of L1 and L2.
        """

        new_states = set()
        new_input_symbols = self.input_symbols | other.input_symbols
        new_transitions = dict()
        new_initial_state = (self.initial_state, other.initial_state)

        queue = deque()

        queue.append(new_initial_state)
        new_states.add(new_initial_state)

        while queue:
            curr_state = queue.popleft()
            q_a, q_b = curr_state

            # States we will consider adding to the queue
            next_states_iterables = list()

            # Get transition dict for states in self
            transitions_a = self.transitions.get(q_a, {})
            # Add epsilon transitions for first set of transitions
            epsilon_transitions_a = transitions_a.get('')
            if epsilon_transitions_a is not None:
                state_dict = new_transitions.setdefault(curr_state, dict())
                state_dict.setdefault('', set()).update(product(epsilon_transitions_a, [q_b]))
                next_states_iterables.append(product(epsilon_transitions_a, [q_b]))

            # Get transition dict for states in other
            transitions_b = other.transitions.get(q_b, {})
            # Add epsilon transitions for second set of transitions
            epsilon_transitions_b = transitions_b.get('')
            if epsilon_transitions_b is not None:
                state_dict = new_transitions.setdefault(curr_state, dict())
                state_dict.setdefault('', set()).update(product([q_a], epsilon_transitions_b))
                next_states_iterables.append(product([q_a], epsilon_transitions_b))

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state, dict())
                    state_dict.setdefault(symbol, set()).update(product(end_states_a, end_states_b))
                    next_states_iterables.append(product(end_states_a, end_states_b))

            # Finally, try visiting every state we found.
            for product_state in chain.from_iterable(next_states_iterables):
                if product_state not in new_states:
                    new_states.add(product_state)
                    queue.append(product_state)

        new_final_states = frozenset(
            (state_a, state_b)
            for (state_a, state_b) in new_states
            if state_a in self.final_states and state_b in other.final_states
        )

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def shuffle_product(self, other):
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the shuffle of L1 and L2.
        """
        if not isinstance(other, NFA):
            raise TypeError(f"other must be an NFA, not {other.__class__.__name__}")

        new_input_symbols = self.input_symbols | other.input_symbols
        new_initial_state = (self.initial_state, other.initial_state)
        new_states = frozenset(product(self.states, other.states))

        new_transitions = dict()

        for curr_state in new_states:
            state_dict = new_transitions.setdefault(curr_state, dict())
            q_a, q_b = curr_state

            transitions_a = self.transitions.get(q_a, dict())
            for symbol, end_states in transitions_a.items():
                state_dict.setdefault(symbol, set()).update(product(end_states, [q_b]))

            transitions_b = other.transitions.get(q_b, dict())
            for symbol, end_states in transitions_b.items():
                state_dict.setdefault(symbol, set()).update(product([q_a], end_states))

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=frozenset(product(self.final_states, other.final_states))
        )

    def right_quotient(self, other):
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the right quotient of L1 with respect to L2 (L1 / L2).

        Construction is based off of the one described here:
        https://cs.stackexchange.com/a/102043
        """

        if not isinstance(other, NFA):
            raise TypeError(f"other must be an NFA, not {other.__class__.__name__}")

        # First, eliminate lambdas because they cause problems with this algorithm
        self_reachable_states, self_new_transitions, self_reachable_final_states = self._eliminate_lambda()
        other_reachable_states, other_new_transitions, other_reachable_final_states = other._eliminate_lambda()

        new_input_symbols = self.input_symbols | other.input_symbols
        new_initial_state = (self.initial_state, other.initial_state, False)
        new_final_states = frozenset(product(self_reachable_final_states, other_reachable_final_states, [True]))
        new_states = frozenset(chain(
            product(self_reachable_states, [other.initial_state], [False]),
            product(self_reachable_states, other_reachable_states, [True])
        ))

        new_transitions = dict()

        # Populate transitions for before reading the suffix
        for state in self_reachable_states:
            new_state = (state, other.initial_state, False)
            new_state_dict = new_transitions.setdefault(new_state, dict())
            old_transitions_dict = self_new_transitions.get(state)

            if old_transitions_dict:
                for symbol, end_states in old_transitions_dict.items():
                    new_state_dict[symbol] = {
                        (end_state, other.initial_state, False)
                        for end_state in end_states
                    }

            new_state_dict[''] = {(state, other.initial_state, True)}

        # Start reading after the suffix
        for q_a, q_b in product(self_reachable_states, other_reachable_states):
            curr_state = (q_a, q_b, True)

            transitions_a = self_new_transitions.get(q_a, {})
            transitions_b = other_new_transitions.get(q_b, {})

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state, dict())
                    state_dict.setdefault('', set()).update(product(end_states_a, end_states_b, [True]))

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def left_quotient(self, other):
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the left quotient of L1 with respect to L2 (L2 \\ L1).

        Construction is based off of the one described here:
        https://cs.stackexchange.com/a/102043
        """

        if not isinstance(other, NFA):
            raise TypeError(f"other must be an NFA, not {other.__class__.__name__}")

        # First, eliminate lambdas because they cause problems with this algorithm
        self_reachable_states, self_new_transitions, self_reachable_final_states = self._eliminate_lambda()
        other_reachable_states, other_new_transitions, other_reachable_final_states = other._eliminate_lambda()

        new_input_symbols = self.input_symbols | other.input_symbols
        new_initial_state = (self.initial_state, other.initial_state, False)
        new_final_states = frozenset(product(self_reachable_final_states, other_reachable_final_states, [True]))
        new_states = frozenset(chain(
            product(self_reachable_states, other_reachable_states, [False]),
            product(self_reachable_states, other_reachable_final_states, [True])
        ))

        new_transitions = dict()

        # Start reading the prefix
        for q_a, q_b in product(self_reachable_states, other_reachable_states):
            curr_state = (q_a, q_b, False)

            transitions_a = self_new_transitions.get(q_a, {})
            transitions_b = other_new_transitions.get(q_b, {})

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state, dict())
                    state_dict.setdefault('', set()).update(product(end_states_a, end_states_b, [False]))

            # Add lambda transition from final state, flipping third entry to true
            if q_b in other_reachable_final_states:
                state_dict = new_transitions.setdefault(curr_state, dict())
                state_dict.setdefault('', set()).update({(q_a, q_b, True)})

        # Populate transitions for after reading the prefix
        for state_a, state_b in product(self_reachable_states, other_reachable_final_states):
            new_state = (state_a, state_b, True)
            new_state_dict = new_transitions.setdefault(new_state, dict())
            old_transitions_dict = self_new_transitions.get(state_a)

            if old_transitions_dict:
                for symbol, end_states in old_transitions_dict.items():
                    new_state_dict[symbol] = {
                        (end_state, state_b, True)
                        for end_state in end_states
                    }

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def show_diagram(self, path=None):
        """
            Creates the graph associated with this DFA
        """
        # Nodes are set of states

        graph = Dot(graph_type='digraph', rankdir='LR')
        nodes = {}
        for state in self.states:
            if state == self.initial_state:
                # color start state with green
                if state in self.final_states:
                    initial_state_node = Node(
                        state,
                        style='filled',
                        peripheries=2,
                        fillcolor='#66cc33')
                else:
                    initial_state_node = Node(
                        state, style='filled', fillcolor='#66cc33')
                nodes[state] = initial_state_node
                graph.add_node(initial_state_node)
            else:
                if state in self.final_states:
                    state_node = Node(state, peripheries=2)
                else:
                    state_node = Node(state)
                nodes[state] = state_node
                graph.add_node(state_node)
        # adding edges
        for from_state, lookup in self.transitions.items():
            for to_label, to_states in lookup.items():
                for to_state in to_states:
                    graph.add_edge(Edge(
                        nodes[from_state],
                        nodes[to_state],
                        label=to_label
                    ))
        if path:
            graph.write_png(path)
        return graph

    @staticmethod
    def _load_new_transition_dict(state_map_dict,
                                  old_transition_dict,
                                  new_transition_dict):
        """
        Load the new_transition_dict with the old transitions corresponding to
        the given state_map_dict.
        """

        for state_a, transitions in old_transition_dict.items():
            for symbol, states in transitions.items():
                new_transition_dict[state_map_dict[state_a]][symbol] = {
                    state_map_dict[state_b] for state_b in states
                }

    @staticmethod
    def _add_new_state(state_set, start=0):
        """Adds new state to the state set and returns it"""
        new_state = start
        while new_state in state_set:
            new_state += 1

        state_set.add(new_state)

        return new_state

    def __eq__(self, other):
        """
        Return True if two NFAs are equivalent. Uses an optimized version of
        the extended Hopcroft-Karp algorithm (HKe). See
        https://arxiv.org/abs/0907.5058
        """

        # Must be another NFA and have equal alphabets
        if not isinstance(other, NFA) or self.input_symbols != other.input_symbols:
            return NotImplemented

        operand_nfas = (self, other)
        initial_state_a = (self._lambda_closures[self.initial_state], 0)
        initial_state_b = (other._lambda_closures[other.initial_state], 1)

        def is_final_state(states_pair):
            states, operand_index = states_pair
            nfa = operand_nfas[operand_index]
            # If at least one of the current states is a final state, the
            # condition should satisfy
            return any(
                len(nfa.final_states & nfa._lambda_closures[state]) > 0
                for state in states
            )

        def transition(states_pair, symbol):
            states, operand_index = states_pair
            return (
                operand_nfas[operand_index]._get_next_current_states(
                    states, symbol),
                operand_index
            )

        # Get data structures
        state_sets = nx.utils.union_find.UnionFind([initial_state_a, initial_state_b])
        pair_stack = deque()

        # Do union find
        state_sets.union(initial_state_a, initial_state_b)
        pair_stack.append((initial_state_a, initial_state_b))

        while pair_stack:
            q_a, q_b = pair_stack.pop()

            if is_final_state(q_a) ^ is_final_state(q_b):
                return False

            for symbol in self.input_symbols:
                r_1 = state_sets[transition(q_a, symbol)]
                r_2 = state_sets[transition(q_b, symbol)]

                if r_1 != r_2:
                    state_sets.union(r_1, r_2)
                    pair_stack.append((r_1, r_2))

        return True

    @classmethod
    def edit_distance(cls, input_symbols, reference_str, max_edit_distance, *,
                      insertion=True, deletion=True, substitution=True):
        """
        Constructs the Levenshtein NFA for the given reference_str and
        given Levenshtein distance. This NFA recognizes strings within the given
        Levenshtein distance (commonly called edit distance) of the reference_str.
        Parameters control which error types the NFA will recognize (insertions,
        deletions, or substitutions). At least one error type must be set.

        If insertion and deletion are False and substitution is True,
        then this is the same as Hamming distance.

        If insertion and deletion are True and substitution is False,
        then this is the same as LCS distance.

        insertion, deletion, and substitution all default to True.

        Code adapted from: http://blog.notdot.net/2010/07/Damn-Cool-Algorithms-Levenshtein-Automata
        """
        if max_edit_distance < 0:
            raise ValueError("max_edit_distance must be greater than zero")
        if not (insertion or deletion or substitution):
            raise ValueError("At least one of insertion, deletion, or substitution must be enabled.")

        states = frozenset(product(range(len(reference_str)+1), range(max_edit_distance+1)))

        transitions = dict()
        final_states = set()

        def add_transition(start_state_dict, end_state, symbol):
            """Add transition between start and end state on symbol"""
            char_transitions = start_state_dict.setdefault(symbol, set())
            char_transitions.add(end_state)

        def add_any_transition(start_state_dict, end_state):
            """Add transition on all symbols between start and end state"""
            for symbol in input_symbols:
                add_transition(start_state_dict, end_state, symbol)

        for i, chr in enumerate(reference_str):
            for e in range(max_edit_distance + 1):
                state_transition_dict = transitions.setdefault((i, e), dict())

                # Correct character
                add_transition(state_transition_dict, (i + 1, e), chr)
                if e < max_edit_distance:
                    if insertion:
                        # Insertion
                        add_any_transition(state_transition_dict, (i, e + 1))

                    if deletion:
                        # Deletion
                        add_transition(state_transition_dict, (i + 1, e + 1), '')

                    if substitution:
                        # Substitution
                        add_any_transition(state_transition_dict, (i + 1, e + 1))

        for e in range(max_edit_distance + 1):
            state_transition_dict = transitions.setdefault((len(reference_str), e), dict())
            if e < max_edit_distance and insertion:
                add_any_transition(state_transition_dict, (len(reference_str), e + 1))

            final_states.add((len(reference_str), e))

        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=(0, 0),
            final_states=final_states
        )



#!/usr/bin/env python3
"""Classes and methods for working with multitape nondeterministic Turing
machines."""

from collections import deque

import automata.base.exceptions as exceptions
import automata.tm.exceptions as tm_exceptions
import automata.tm.ntm as ntm
import automata.tm.tm as tm
from automata.tm.configuration import MTMConfiguration, TMConfiguration
from automata.tm.tape import TMTape


class MNTM(ntm.NTM):
    """A multitape nondeterministic Turing machine."""

    __slots__ = ('states', 'input_symbols', 'tape_symbols', 'n_tapes',
                 'transitions', 'initial_state', 'blank_symbol', 'final_states')

    def __init__(self, *, states, input_symbols, tape_symbols, n_tapes,
                 transitions, initial_state, blank_symbol, final_states):
        """Initialize a complete Turing machine."""
        super(tm.TM, self).__init__(
            states=states,
            input_symbols=input_symbols,
            tape_symbols=tape_symbols,
            transitions=transitions,
            initial_state=initial_state,
            blank_symbol=blank_symbol,
            final_states=final_states,
            n_tapes=n_tapes
        )

    def _validate_transition_symbols(self, state, paths):
        for tape_symbol in [tape_symbol
                            for symbol in paths.keys() for tape_symbol in
                            symbol]:
            if tape_symbol not in self.tape_symbols:
                raise exceptions.InvalidSymbolError(
                    'transition symbol {} for state {} is not valid'.format(
                        tape_symbol, state
                    )
                )

    def _validate_transition_state(self, transition_state):
        if transition_state not in self.states:
            raise exceptions.InvalidStateError(
                'transition state is not valid ({})'.format(transition_state)
            )

    def _validate_transition_results(self, paths):
        for results in paths.values():
            for result in results:
                state, moves = result
                for move in moves:
                    symbol, direction = move
                    possible_result = (state, symbol, direction)
                    self._validate_transition_result(possible_result)

    def _validate_tapes_consistency(self):
        for state in self.transitions:
            for read_tape_symbols in self.transitions[state]:
                if len(read_tape_symbols) != self.n_tapes:
                    error = (
                        'tapes symbols {} inconsistent with the number of '
                        'tapes defined. Expected {} symbols, got {}'
                    ).format(
                        read_tape_symbols,
                        self.n_tapes,
                        len(read_tape_symbols)
                    )
                    raise tm_exceptions.InconsistentTapesException(
                        error
                    )
                for transition in self.transitions[state][read_tape_symbols]:
                    _, moves = transition
                    if len(moves) != self.n_tapes:
                        error = (
                            'transition {} has inconsistent operations on '
                            'tapes. Expected {} write/move operations, '
                            'got {}'
                        ).format(
                            transition,
                            self.n_tapes,
                            len(moves)
                        )
                        raise tm_exceptions.InconsistentTapesException(
                            error
                        )

    def validate(self):
        """Return True if this MNTM is internally consistent."""
        super().validate()
        self._validate_tapes_consistency()
        return True

    def _get_tapes_for_input_str(self, input_str):
        """Produce a new list of tapes based on an input string to be read."""
        return [
            # Input is saved on first tape
            TMTape(input_str, blank_symbol=self.blank_symbol),
            # The rest of the tapes have blanks
            *(
                TMTape(self.blank_symbol,
                       blank_symbol=self.blank_symbol,
                       current_position=0)
                for _ in range(self.n_tapes - 1)
            )
        ]

    def _read_current_tape_symbols(self, tapes):
        """Reads the current tape symbols in each of the tapes and their
        corresponding heads."""
        return tuple(tape.read_symbol() for tape in tapes)

    def _get_transition(self, current_state, tapes):
        """Get the transition tuple for the given state and tape symbols in
        each tape."""
        current_tape_symbols = self._read_current_tape_symbols(tapes)
        if current_state in self.transitions and current_tape_symbols in \
                self.transitions[current_state]:
            return self.transitions[current_state][
                self._read_current_tape_symbols(tapes)
            ]
        else:
            return None

    def _get_next_configuration(self, transition, current_tapes):
        """Advances to the next configuration."""
        current_state, moves = transition
        tapes = current_tapes.copy()
        for i, move in enumerate(moves):
            symbol, direction = move
            tapes[i] = tapes[i].write_symbol(symbol)
            tapes[i] = tapes[i].move(direction)

        return MTMConfiguration(state=current_state, tapes=tapes)

    def _has_accepted(self, current_config):
        return current_config.state in self.final_states

    def read_input_stepwise(self, input_str):
        """Checks if the given string is accepted by this Turing machine,
        using a BFS of every possible configuration from each configuration.
        Yields the current configuration of the machine at each step.
        """
        tapes = self._get_tapes_for_input_str(input_str)
        queue = deque([(
            MTMConfiguration(state=self.initial_state, tapes=tapes[:])
        )])
        while len(queue) > 0:
            current_config = queue.popleft()
            yield {MTMConfiguration(current_config.state, tuple(current_config.tapes))}

            possible_transitions = self._get_transition(current_config.state, current_config.tapes)
            if possible_transitions is None:
                if self._has_accepted(current_config):
                    return {MTMConfiguration(current_config.state,
                                             tuple(current_config.tapes))}
            else:
                for transition in possible_transitions[1:]:
                    queue.append(self._get_next_configuration(
                        transition,
                        current_tapes=current_config.tapes))

                queue.append(self._get_next_configuration(
                    possible_transitions[0],
                    current_tapes=current_config.tapes))

        raise exceptions.RejectionException(
            'the multitape MNTM did not reach an accepting configuration'
        )

    @staticmethod
    def _read_extended_tape(tape: str, head_symbol: str = '^',
                            tape_separator_symbol: str = '_'):
        """Returns a tuple with the symbols extracted from the given
        tape, that are the virtual heads for their corresponding
        virtual tape."""
        virtual_heads = []
        heads_found = 0
        separators_found = 0
        for i, symbol in enumerate(tape):
            if symbol == head_symbol:
                if i - 1 < 0:
                    raise tm_exceptions.MalformedExtendedTapeError(
                        'head symbol was found on leftmost end of the '
                        'extended tape'
                    )
                else:
                    previous_symbol = tape[i - 1]
                    virtual_heads.append(previous_symbol)
                    heads_found += 1
            elif symbol == tape_separator_symbol:
                if heads_found == 0:
                    raise tm_exceptions.MalformedExtendedTapeError(
                        'no head symbol found on one of the virtual tapes'
                    )
                elif heads_found > 1:
                    raise tm_exceptions.MalformedExtendedTapeError(
                        'more than one head symbol found on one of the '
                        'virtual tapes'
                    )
                else:
                    heads_found = 0
                    separators_found += 1

        if len(virtual_heads) != separators_found:
            raise tm_exceptions.MalformedExtendedTapeError(
                'there must be 1 virtual head for every tape separator symbol'
            )

        return tuple(virtual_heads)

    def read_input_as_ntm(self, input_str):
        """Simulates the machine as a single-tape Turing machine.
        Yields the configuration at each step."""
        tapes = self._get_tapes_for_input_str(input_str)
        head_symbol = '^'
        tape_separator_symbol = '_'
        extended_tape = ''
        tapes_copy = tapes.copy()
        for tape_copy in tapes_copy:
            tape_str = tape_copy.get_symbols_as_str()
            extended_tape += tape_str[0] + head_symbol + \
                tape_str[1:] + tape_separator_symbol

        current_state = self.initial_state
        yield {
            TMConfiguration(current_state,
                            TMTape(extended_tape,
                                   blank_symbol=self.blank_symbol,
                                   current_position=0)
                            )
        }

        # If the machine has not reached an accepting state.
        while current_state not in self.final_states:
            i = 0  # current position
            virtual_heads = self._read_extended_tape(extended_tape,
                                                     head_symbol,
                                                     tape_separator_symbol)
            try:
                next_config = self.transitions[current_state][virtual_heads]
            except KeyError:
                raise exceptions.RejectionException(
                    'the multitape NTM did not reach an accepting '
                    'configuration')
            next_state, moves = next_config[0]
            for move in moves:
                new_head, direction = move
                executing_changes = True

                while executing_changes:
                    if extended_tape[i] == head_symbol:
                        # Head has been found (previous symbol is the head).
                        # This replaces the previous symbol with the new_head.
                        extended_tape = (
                            extended_tape[: i - 1] +
                            new_head + extended_tape[i:]
                        )
                        extended_tape = extended_tape[:i] + \
                            '' + extended_tape[i + 1:]

                        # After replacing, the machine must change the
                        # position of the virtual head of the current virtual
                        # tape.
                        if direction == 'R':
                            i += 1
                        elif direction == 'L':
                            i -= 1
                        else:  # direction == 'N'
                            i += 0

                        if extended_tape[i - 1] == tape_separator_symbol:
                            i -= 1
                            extended_tape = extended_tape[:i] + \
                                self.blank_symbol + \
                                head_symbol + extended_tape[i:]

                            i += 1
                        else:
                            extended_tape = (
                                extended_tape[:i] +
                                head_symbol + extended_tape[i:]
                            )

                    elif extended_tape[i] == tape_separator_symbol:
                        executing_changes = False

                    i += 1

            current_state = next_state
            yield {
                TMConfiguration(current_state,
                                TMTape(extended_tape,
                                       blank_symbol=self.blank_symbol,
                                       current_position=i-1)
                                )
            }



#!/usr/bin/env python3
"""Classes for working with all automata, including Turing machines."""

import abc

from frozendict import frozendict

import automata.base.config as global_config
import automata.base.exceptions as exceptions
from automata.base.utils import freezeValue


class Automaton(metaclass=abc.ABCMeta):
    """An abstract base class for all automata, including Turing machines."""

    def __init__(self, **kwargs):
        for attr_name, attr_value in kwargs.items():
            object.__setattr__(self, attr_name, freezeValue(attr_value))
        self.__post_init__()

    def __post_init__(self):
        if global_config.should_validate_automata:
            self.validate()

    @abc.abstractmethod
    def validate(self):
        """Return True if this automaton is internally consistent."""
        raise NotImplementedError

    def __setattr__(self, name, value):
        """Set custom setattr to make class immutable."""
        raise AttributeError(f'This {type(self).__name__} is immutable')

    def __delattr__(self, name):
        """Set custom delattr to make class immutable."""
        raise AttributeError(f'This {type(self).__name__} is immutable')

    @abc.abstractmethod
    def read_input_stepwise(self, input_str):
        """Return a generator that yields each step while reading input."""
        raise NotImplementedError

    def read_input(self, input_str):
        """
        Check if the given string is accepted by this automaton.

        Return the automaton's final configuration if this string is valid.
        """
        # "Fast-forward" generator to get its final value
        for config in self.read_input_stepwise(input_str):
            pass
        return config

    def accepts_input(self, input_str):
        """Return True if this automaton accepts the given input."""
        try:
            self.read_input(input_str)
            return True
        except exceptions.RejectionException:
            return False

    def _validate_initial_state(self):
        """Raise an error if the initial state is invalid."""
        if self.initial_state not in self.states:
            raise exceptions.InvalidStateError(
                '{} is not a valid initial state'.format(self.initial_state))

    def _validate_initial_state_transitions(self):
        """Raise an error if the initial state has no transitions defined."""
        if self.initial_state not in self.transitions and len(self.states) > 1:
            raise exceptions.MissingStateError(
                'initial state {} has no transitions defined'.format(
                    self.initial_state))

    def _validate_final_states(self):
        """Raise an error if any final states are invalid."""
        invalid_states = self.final_states - self.states
        if invalid_states:
            raise exceptions.InvalidStateError(
                'final states are not valid ({})'.format(
                    ', '.join(str(state) for state in invalid_states)))

    @property
    def input_parameters(self):
        """Return the public attributes for this automaton."""
        return {attr_name: getattr(self, attr_name)
                for attr_name in self.__slots__
                if not attr_name.startswith('_')}

    def copy(self):
        """Create a deep copy of the automaton."""
        return self.__class__(**self.input_parameters)

    # Format the given value for string output via repr() or str(); this exists for the purpose of displaying

    def _get_repr_friendly_value(self, value):
        """
        A helper function to convert the given value / structure into a fully
        mutable one by recursively processing said structure and any of its
        members, unfreezing them along the way
        """
        if isinstance(value, frozenset):
            return {self._get_repr_friendly_value(element)
                    for element in value}
        elif isinstance(value, frozendict):
            return {
                dict_key: self._get_repr_friendly_value(dict_value)
                for dict_key, dict_value in value.items()
            }
        else:
            return value

    def __repr__(self):
        """Return a string representation of the automaton."""
        values = ', '.join(
            f'{attr_name}={self._get_repr_friendly_value(attr_value)!r}'
            for attr_name, attr_value in self.input_parameters.items())
        return f'{self.__class__.__qualname__}({values})'

    def __contains__(self, input_str):
        """Returns whether the word is accepted by the automaton."""
        return self.accepts_input(input_str)



#!/usr/bin/env python3
"""Classes and methods for working with deterministic finite automata."""

from collections import defaultdict, deque
from itertools import chain, count
from random import Random

import networkx as nx
from pydot import Dot, Edge, Node

import automata.base.exceptions as exceptions
import automata.fa.fa as fa
from automata.base.utils import PartitionRefinement


class DFA(fa.FA):
    """A deterministic finite automaton."""

    __slots__ = ('states', 'input_symbols', 'transitions',
                 'initial_state', 'final_states', 'allow_partial')

    def __init__(self, *, states, input_symbols, transitions,
                 initial_state, final_states, allow_partial=False):
        """Initialize a complete DFA."""
        super().__init__(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=initial_state,
            final_states=final_states,
            allow_partial=allow_partial
        )
        object.__setattr__(self, '_word_cache', [])
        object.__setattr__(self, '_count_cache', [])

    def __eq__(self, other):
        """
        Return True if two DFAs are equivalent. Uses an optimized version of
        the Hopcroft-Karp algorithm. See https://arxiv.org/abs/0907.5058
        """

        # Must be another DFA and have equal alphabets
        if not isinstance(other, DFA) or self.input_symbols != other.input_symbols:
            return NotImplemented

        operand_dfas = (self, other)
        initial_state_a = (self.initial_state, 0)
        initial_state_b = (other.initial_state, 1)

        def is_final_state(state_pair):
            state, operand_index = state_pair
            return state in operand_dfas[operand_index].final_states

        def transition(state_pair, symbol):
            state, operand_index = state_pair
            return (
                operand_dfas[operand_index]._get_next_current_state(
                    state, symbol),
                operand_index
            )

        # Get data structures
        state_sets = nx.utils.union_find.UnionFind((initial_state_a, initial_state_b))
        pair_stack = deque()

        # Do union find
        state_sets.union(initial_state_a, initial_state_b)
        pair_stack.append((initial_state_a, initial_state_b))

        while pair_stack:
            q_a, q_b = pair_stack.pop()

            if is_final_state(q_a) ^ is_final_state(q_b):
                return False

            for symbol in self.input_symbols:

                r_1 = state_sets[transition(q_a, symbol)]
                r_2 = state_sets[transition(q_b, symbol)]

                if r_1 != r_2:
                    state_sets.union(r_1, r_2)
                    pair_stack.append((r_1, r_2))

        return True

    def __le__(self, other):
        """Return True if this DFA is a subset of (or equal to) another DFA."""
        if isinstance(other, DFA):
            return self.issubset(other)
        else:
            return NotImplemented

    def __ge__(self, other):
        """Return True if this DFA is a superset of another DFA."""
        if isinstance(other, DFA):
            return self.issuperset(other)
        else:
            return NotImplemented

    def __lt__(self, other):
        """Return True if this DFA is a strict subset of another DFA."""
        if isinstance(other, DFA):
            return self <= other and self != other
        else:
            return NotImplemented

    def __gt__(self, other):
        """Return True if this DFA is a strict superset of another DFA."""
        if isinstance(other, DFA):
            return self >= other and self != other
        else:
            return NotImplemented

    def __sub__(self, other):
        """Return a DFA that is the difference of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.difference(other)
        else:
            return NotImplemented

    def __or__(self, other):
        """Return the union of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.union(other)
        else:
            return NotImplemented

    def __and__(self, other):
        """Return the intersection of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.intersection(other)
        else:
            return NotImplemented

    def __xor__(self, other):
        """Return the symmetric difference of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.symmetric_difference(other)
        else:
            return NotImplemented

    def __invert__(self):
        """Return the complement of this DFA and another DFA."""
        return self.complement()

    def __iter__(self):
        """
        Iterates through all words in the language represented by the DFA.
        The words are ordered first by length and then by the order of the input symbol set.
        """
        i = self.minimum_word_length()
        limit = self.maximum_word_length()
        while limit is None or i <= limit:
            yield from self.words_of_length(i)
            i += 1

    def __len__(self):
        """Returns the cardinality of the language represented by the DFA."""
        return self.cardinality()

    def _validate_transition_missing_symbols(self, start_state, paths):
        """Raise an error if the transition input_symbols are missing."""
        if self.allow_partial:
            return
        for input_symbol in self.input_symbols:
            if input_symbol not in paths:
                raise exceptions.MissingSymbolError(
                    'state {} is missing transitions for symbol {}'.format(
                        start_state, input_symbol))

    def _validate_transition_invalid_symbols(self, start_state, paths):
        """Raise an error if transition input symbols are invalid."""
        for input_symbol in paths.keys():
            if input_symbol not in self.input_symbols:
                raise exceptions.InvalidSymbolError(
                    'state {} has invalid transition symbol {}'.format(
                        start_state, input_symbol))

    def _validate_transition_start_states(self):
        """Raise an error if transition start states are missing."""
        if self.allow_partial:
            return
        for state in self.states:
            if state not in self.transitions:
                raise exceptions.MissingStateError(
                    'transition start state {} is missing'.format(
                        state))

    def _validate_transition_end_states(self, start_state, paths):
        """Raise an error if transition end states are invalid."""
        for end_state in paths.values():
            if end_state not in self.states:
                raise exceptions.InvalidStateError(
                    'end state {} for transition on {} is not valid'.format(
                        end_state, start_state))

    def _validate_transitions(self, start_state, paths):
        """Raise an error if transitions are missing or invalid."""
        self._validate_transition_missing_symbols(start_state, paths)
        self._validate_transition_invalid_symbols(start_state, paths)
        self._validate_transition_end_states(start_state, paths)

    def validate(self):
        """Return True if this DFA is internally consistent."""
        self._validate_transition_start_states()
        for start_state, paths in self.transitions.items():
            self._validate_transitions(start_state, paths)
        self._validate_initial_state()
        self._validate_final_states()
        return True

    def _get_next_current_state(self, current_state, input_symbol):
        """
        Follow the transition for the given input symbol on the current state.

        Raise an error if the transition does not exist.
        """
        if current_state is not None and input_symbol in self.transitions[current_state]:
            return self.transitions[current_state][input_symbol]
        return None

    def _check_for_input_rejection(self, current_state):
        """Raise an error if the given config indicates rejected input."""
        if current_state not in self.final_states:
            raise exceptions.RejectionException(
                'the DFA stopped on a non-final state ({})'.format(
                    current_state))

    def read_input_stepwise(self, input_str, ignore_rejection=False):
        """
        Check if the given string is accepted by this DFA.

        Yield the current configuration of the DFA at each step.
        """
        current_state = self.initial_state

        yield current_state
        for input_symbol in input_str:
            current_state = self._get_next_current_state(
                current_state, input_symbol)
            yield current_state

        if not ignore_rejection:
            self._check_for_input_rejection(current_state)

    def _get_digraph(self):
        """Return a digraph corresponding to this DFA with transition symbols ignored"""
        return nx.DiGraph([
            (start_state, end_state)
            for start_state, transition in self.transitions.items()
            for end_state in transition.values()
        ])

    def _compute_reachable_states(self):
        """Compute the states which are reachable from the initial state."""
        visited_set = set()
        queue = deque()

        queue.append(self.initial_state)
        visited_set.add(self.initial_state)

        while queue:
            state = queue.popleft()

            for next_state in self.transitions[state].values():
                if next_state not in visited_set:
                    visited_set.add(next_state)
                    queue.append(next_state)

        return visited_set

    def minify(self, retain_names=False):
        """
        Create a minimal DFA which accepts the same inputs as this DFA.

        First, non-reachable states are removed.
        Then, similiar states are merged using Hopcroft's Algorithm.
            retain_names: If True, merged states retain names.
                          If False, new states will be named 0, ..., n-1.
        """

        # Compute reachable states and final states
        reachable_states = self._compute_reachable_states()
        reachable_final_states = self.final_states & reachable_states

        return self._minify(
            reachable_states=reachable_states,
            input_symbols=self.input_symbols,
            transitions=self.transitions,
            initial_state=self.initial_state,
            reachable_final_states=reachable_final_states,
            retain_names=retain_names)

    @classmethod
    def _minify(cls, *, reachable_states, input_symbols, transitions, initial_state,
                reachable_final_states, retain_names):
        """Minify helper function. DFA data passed in must have no unreachable states."""

        # First, assemble backmap and equivalence class data structure
        eq_classes = PartitionRefinement(reachable_states)
        refinement = eq_classes.refine(reachable_final_states)

        final_states_id = refinement[0][0] if refinement else next(iter(eq_classes.get_set_ids()))

        transition_back_map = {
            symbol: {
                end_state: list()
                for end_state in reachable_states
            }
            for symbol in input_symbols
        }

        for start_state, path in transitions.items():
            if start_state in reachable_states:
                for symbol, end_state in path.items():
                    transition_back_map[symbol][end_state].append(start_state)

        origin_dicts = tuple(transition_back_map.values())
        processing = {final_states_id}

        while processing:
            # Save a copy of the set, since it could get modified while executing
            active_state = tuple(eq_classes.get_set_by_id(processing.pop()))
            for origin_dict in origin_dicts:
                states_that_move_into_active_state = chain.from_iterable(
                    origin_dict[end_state] for end_state in active_state
                )

                # Refine set partition by states moving into current active one
                new_eq_class_pairs = eq_classes.refine(states_that_move_into_active_state)

                for (YintX_id, YdiffX_id) in new_eq_class_pairs:
                    # Only adding one id to processing, since the other is already there
                    if YdiffX_id in processing:
                        processing.add(YintX_id)
                    else:
                        if len(eq_classes.get_set_by_id(YintX_id)) <= len(eq_classes.get_set_by_id(YdiffX_id)):
                            processing.add(YintX_id)
                        else:
                            processing.add(YdiffX_id)

        # now eq_classes are good to go, make them a list for ordering
        eq_class_name_pairs = (
            [(frozenset(eq), eq) for eq in eq_classes.get_sets()] if retain_names else
            list(enumerate(eq_classes.get_sets()))
        )

        # need a backmap to prevent constant calls to index
        back_map = {
            state: name
            for name, eq in eq_class_name_pairs
            for state in eq
        }

        new_input_symbols = input_symbols
        new_states = frozenset(back_map.values())
        new_initial_state = back_map[initial_state]
        new_final_states = frozenset(back_map[acc] for acc in reachable_final_states)
        new_transitions = {
            name: {
                letter: back_map[transitions[next(iter(eq))][letter]]
                for letter in input_symbols
            }
            for name, eq in eq_class_name_pairs
        }

        return cls(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states,
        )

    def union(self, other, *, retain_names=False, minify=True):
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the union of L1 and L2.
        """

        new_states, new_transitions, new_initial_state = self._cross_product(other, False)

        new_final_states = frozenset(
            (state_a, state_b)
            for state_a, state_b in new_states
            if state_a in self.final_states or state_b in other.final_states
        )

        if minify:
            return self._minify(
                reachable_states=new_states,
                input_symbols=self.input_symbols,
                transitions=new_transitions,
                initial_state=new_initial_state,
                reachable_final_states=new_final_states,
                retain_names=retain_names)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def intersection(self, other, *, retain_names=False, minify=True):
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the intersection of L1 and L2.
        """

        new_states, new_transitions, new_initial_state = self._cross_product(other, False)

        new_final_states = frozenset(
            (state_a, state_b)
            for state_a, state_b in new_states
            if state_a in self.final_states and state_b in other.final_states
        )

        if minify:
            return self._minify(
                reachable_states=new_states,
                input_symbols=self.input_symbols,
                transitions=new_transitions,
                initial_state=new_initial_state,
                reachable_final_states=new_final_states,
                retain_names=retain_names)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def difference(self, other, *, retain_names=False, minify=True):
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the difference of L1 and L2.
        """

        new_states, new_transitions, new_initial_state = self._cross_product(other, False)

        new_final_states = frozenset(
            (state_a, state_b)
            for state_a, state_b in new_states
            if state_a in self.final_states and state_b not in other.final_states
        )

        if minify:
            return self._minify(
                reachable_states=new_states,
                input_symbols=self.input_symbols,
                transitions=new_transitions,
                initial_state=new_initial_state,
                reachable_final_states=new_final_states,
                retain_names=retain_names)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def symmetric_difference(self, other, *, retain_names=False, minify=True):
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the symmetric difference of L1 and L2.
        """

        new_states, new_transitions, new_initial_state = self._cross_product(other, False)

        new_final_states = frozenset(
            (state_a, state_b)
            for state_a, state_b in new_states
            if (state_a in self.final_states) ^ (state_b in other.final_states)
        )

        if minify:
            return self._minify(
                reachable_states=new_states,
                input_symbols=self.input_symbols,
                transitions=new_transitions,
                initial_state=new_initial_state,
                reachable_final_states=new_final_states,
                retain_names=retain_names)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states
        )

    def complement(self, *, retain_names=False, minify=True):
        """Return the complement of this DFA."""

        if minify:
            reachable_states = self._compute_reachable_states()
            reachable_final_states = self.final_states & reachable_states

            return self._minify(
                reachable_states=reachable_states,
                input_symbols=self.input_symbols,
                transitions=self.transitions,
                initial_state=self.initial_state,
                reachable_final_states=reachable_states - reachable_final_states,
                retain_names=retain_names)

        return self.__class__(
            states=self.states,
            input_symbols=self.input_symbols,
            transitions=self.transitions,
            initial_state=self.initial_state,
            final_states=self.states - self.final_states,
            allow_partial=self.allow_partial
        )

    def _cross_product(self, other, states_only):
        """
        Get reachable states corresponding to product graph between self and other.
        If states_only is True, only returns states. Otherwise, constructs corresponding
        transitions to the reachable states and new initial state.
        """
        if self.input_symbols != other.input_symbols:
            raise exceptions.SymbolMismatchError('The input symbols between the two given DFAs do not match')

        product_transitions = dict() if not states_only else None

        visited_set = set()
        queue = deque()

        product_initial_state = (self.initial_state, other.initial_state)
        queue.append(product_initial_state)
        visited_set.add(product_initial_state)

        while queue:
            # Get next state in BFS queue
            curr_state = queue.popleft()

            # Add state to the transition dict
            if not states_only:
                state_transitions = product_transitions.setdefault(curr_state, dict())

            # Unpack state and get transitions
            q_a, q_b = curr_state
            transitions_a = self.transitions[q_a]
            transitions_b = other.transitions[q_b]

            for chr in self.input_symbols:
                product_state = (transitions_a[chr], transitions_b[chr])

                if not states_only:
                    state_transitions[chr] = product_state

                # If next state is new, add to queue
                if product_state not in visited_set:
                    visited_set.add(product_state)
                    queue.append(product_state)

        if not states_only:
            return visited_set, product_transitions, product_initial_state

        return visited_set

    def issubset(self, other):
        """Return True if this DFA is a subset of another DFA."""
        for (state_a, state_b) in self._cross_product(other, True):
            # Check for reachable state that is counterexample to subset
            if state_a in self.final_states and state_b not in other.final_states:
                return False

        return True

    def issuperset(self, other):
        """Return True if this DFA is a superset of another DFA."""
        return other.issubset(self)

    def isdisjoint(self, other):
        """Return True if this DFA has no common elements with another DFA."""
        for (state_a, state_b) in self._cross_product(other, True):
            # Check for reachable state that is counterexample to disjointness
            if state_a in self.final_states and state_b in other.final_states:
                return False

        return True

    def isempty(self):
        """Return True if this DFA is completely empty."""
        return len(self._compute_reachable_states() & self.final_states) == 0

    def isfinite(self):
        """
        Returns True if the DFA accepts a finite language, False otherwise.
        """
        try:
            return self.maximum_word_length() is not None
        except exceptions.EmptyLanguageException:
            return True

    def random_word(self, k, *, seed=None):
        self._populate_count_cache_up_to_len(k)
        state = self.initial_state
        if self._count_cache[k][state] == 0:
            raise ValueError(f"Language has no words of length {k}")

        result = []
        rng = Random(seed)
        for remaining in range(k, 0, -1):
            total = self._count_cache[remaining][state]
            choice = rng.randint(0, total-1)
            for symbol, next_state in self.transitions[state].items():  # pragma: no branch
                next_state_count = self._count_cache[remaining - 1][next_state]
                if choice < next_state_count:
                    result.append(symbol)
                    state = next_state
                    break
                choice -= next_state_count

        assert state in self.final_states
        return ''.join(result)

    def predecessor(self, input_str, *, strict=True, key=None):
        """
        Returns the first string accepted by the DFA that comes before
        the input string in lexicographical order.
        Passing in None will generate the lexicographically last word.
        If strict is set to False and input_str is accepted by the DFA then
        it will be returned.
        The value of key can be set to define a custom lexicographical ordering.
        """
        for word in self.predecessors(input_str, strict=strict, key=key):
            return word
        return None

    def predecessors(self, input_str, *, strict=True, key=None):
        """
        Generates all strings that come before the input string
        in lexicographical order.
        Passing in None will generate all words.
        If strict is set to False and input_str is accepted by the DFA then
        it will be included in the output.
        The value of key can be set to define a custom lexicographical ordering.
        Raises an InfiniteLanguageException for infinite languages.
        """
        yield from self.successors(input_str, strict=strict, reverse=True, key=key)

    def successor(self, input_str, *, strict=True, key=None):
        """
        Returns the first string accepted by the DFA that comes after
        the input string in lexicographical order.
        Passing in None will generate the lexicographically first word.
        If strict is set to False and input_str is accepted by the DFA then
        it will be returned.
        The value of key can be set to define a custom lexicographical ordering.
        """
        for word in self.successors(input_str, strict=strict, key=key):
            return word
        return None

    def successors(self, input_str, *, strict=True, key=None, reverse=False):
        """
        Generates all strings that come after the input string
        in lexicographical order.
        Passing in None will generate all words.
        If strict is set to False and input_str is accepted by the DFA then
        it will be included in the output.
        If reverse is set to True then predecessors will be generated instead. See the DFA.predecessors method.
        The value of key can be set to define a custom lexicographical ordering.
        """
        # A predecessor for a finite string may be infinite but a successor for a finite string is always finite
        if reverse and not self.isfinite():
            raise exceptions.InfiniteLanguageException('Predecessors cannot be computed for infinite languages')
        graph = self._get_digraph()
        coaccessible_nodes = self.final_states.union(*(
            nx.ancestors(graph, state)
            for state in self.final_states
        ))

        # Precomputations and setup
        include_input = not strict
        sorted_symbols = sorted(self.input_symbols, reverse=reverse, key=key)
        symbol_succ = {symbol_a: symbol_b
                       for symbol_a, symbol_b in zip(sorted_symbols, sorted_symbols[1:])}
        symbol_succ[sorted_symbols[-1]] = None
        # Special case for None
        state_stack = deque([self.initial_state]
                            if input_str is None
                            else self.read_input_stepwise(input_str, ignore_rejection=True))
        char_stack = [] if input_str is None else list(input_str)
        first_symbol = sorted_symbols[0]
        # For predecessors we need to special case the input string None
        candidate = None if reverse and input_str is not None else first_symbol
        # If input_str is None then we yield on first value found no matter the value of include_input
        should_yield = include_input or input_str is None

        # Iterative preorder/postorder (depends on reverse) traversal that yields on final states
        while char_stack or candidate is not None:
            state = state_stack[-1]
            # Successors yield here
            if not reverse and should_yield and candidate == first_symbol and state in self.final_states:
                yield ''.join(char_stack)
            candidate_state = None if candidate is None else self._get_next_current_state(state, candidate)
            # Traverse to child if candidate is viable
            if candidate_state in coaccessible_nodes:
                state_stack.append(candidate_state)
                char_stack.append(candidate)
                candidate = first_symbol
            else:
                # Predecessors yield here
                if reverse and should_yield and candidate is None and state in self.final_states:
                    yield ''.join(char_stack)
                # Candidate is None means no more children to explore, so traverse to parent
                if candidate is None:
                    state = state_stack.pop()
                    candidate = char_stack.pop()
                candidate = symbol_succ[candidate]
            should_yield = True
        # Predecessor yields here for empty string
        if reverse and should_yield and candidate is None and state in self.final_states:
            yield ''.join(char_stack)

    def count_words_of_length(self, k):
        """
        Counts words of length `k` accepted by the DFA
        """
        self._populate_count_cache_up_to_len(k)
        return self._count_cache[k][self.initial_state]

    def _populate_count_cache_up_to_len(self, k):
        """
        Populate count cache up to length k
        """
        while len(self._count_cache) <= k:
            i = len(self._count_cache)
            self._count_cache.append(defaultdict(int))
            level = self._count_cache[i]
            if i == 0:
                level.update({state: 1 for state in self.final_states})
            else:
                prev_level = self._count_cache[i-1]
                level.update({
                    state: sum(prev_level[suffix_state] for suffix_state in self.transitions[state].values())
                    for state in self.states
                })

    def words_of_length(self, k):
        """
        Generates all words of size k in the language represented by the DFA
        """
        self._populate_word_cache_up_to_len(k)
        for word in self._word_cache[k][self.initial_state]:
            yield word

    def _populate_word_cache_up_to_len(self, k):
        """
        Populate word cache up to length k
        """
        sorted_symbols = sorted(self.input_symbols)
        while len(self._word_cache) <= k:
            i = len(self._word_cache)
            self._word_cache.append(defaultdict(list))
            level = self._word_cache[i]
            if i == 0:
                level.update({state: [''] for state in self.final_states})
            else:
                prev_level = self._word_cache[i-1]
                level.update({
                    state: [symbol+word
                            for symbol in sorted_symbols
                            for word in prev_level[self.transitions[state][symbol]]]
                    for state in self.states
                })

    def cardinality(self):
        """Returns the cardinality of the language represented by the DFA."""
        try:
            i = self.minimum_word_length()
        except exceptions.EmptyLanguageException:
            return 0
        limit = self.maximum_word_length()
        if limit is None:
            raise exceptions.InfiniteLanguageException("The language represented by the DFA is infinite.")
        return sum(self.count_words_of_length(j) for j in range(i, limit+1))

    def minimum_word_length(self):
        """
        Returns the length of the shortest word in the language represented by the DFA
        """
        queue = deque()
        distances = defaultdict(lambda: None)
        distances[self.initial_state] = 0
        queue.append(self.initial_state)
        while queue:
            state = queue.popleft()
            if state in self.final_states:
                return distances[state]
            for next_state in self.transitions[state].values():
                if distances[next_state] is None:
                    distances[next_state] = distances[state] + 1
                    queue.append(next_state)
        raise exceptions.EmptyLanguageException('The language represented by the DFA is empty')

    def maximum_word_length(self):
        """
        Returns the length of the longest word in the language represented by the DFA
        In the case of infinite languages, `None` is returned
        """
        if self.isempty():
            raise exceptions.EmptyLanguageException('The language represented by the DFA is empty')
        graph = self._get_digraph()

        accessible_nodes = nx.descendants(graph, self.initial_state) | {self.initial_state}

        coaccessible_nodes = self.final_states.union(*(
            nx.ancestors(graph, state)
            for state in self.final_states
        ))

        important_nodes = accessible_nodes.intersection(coaccessible_nodes)
        subgraph = graph.subgraph(important_nodes)
        try:
            return nx.dag_longest_path_length(subgraph)
        except nx.exception.NetworkXUnfeasible:
            return None

    @classmethod
    def from_prefix(cls, input_symbols, prefix, *, contains=True):
        """
        Directly computes the minimal DFA recognizing strings with the
        given prefix.
        If contains is set to False then the complement is constructed instead.
        """
        err_state = -1
        last_state = len(prefix)
        transitions = {i: {symbol: i+1 if symbol == char else err_state
                           for symbol in input_symbols}
                       for i, char in enumerate(prefix)}
        transitions[last_state] = {symbol: last_state for symbol in input_symbols}
        transitions[err_state] = {symbol: err_state for symbol in input_symbols}

        states = frozenset(transitions.keys())
        final_states = {last_state}
        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states if contains else states - final_states
        )

    @classmethod
    def from_suffix(cls, input_symbols, suffix, *, contains=True):
        """
        Directly computes the minimal DFA recognizing strings with the
        given prefix.
        If contains is set to False then the complement is constructed instead.
        """
        return cls.from_substring(input_symbols, suffix, contains=contains, must_be_suffix=True)

    @classmethod
    def from_substring(cls, input_symbols, substring, *, contains=True, must_be_suffix=False):
        """
        Directly computes the minimal DFA recognizing strings containing the
        given substring.
        If contains is set to False then the complement is constructed instead.
        If must_be_suffix is set to True, then the substring must be a suffix instead.
        """
        transitions = {i: dict() for i in range(len(substring))}
        transitions[len(substring)] = {
            symbol: len(substring) for symbol in input_symbols
        }

        # Computing failure function for partial matches as is done in the
        # Knuth-Morris-Pratt string algorithm so we can quickly compute the
        # next state from another state
        kmp_table = [-1 for _ in substring]
        candidate = 0
        for i, char in enumerate(substring):
            if i == 0:
                continue
            elif char == substring[candidate]:
                kmp_table[i] = kmp_table[candidate]
            else:
                kmp_table[i] = candidate
                while candidate >= 0 and char != substring[candidate]:
                    candidate = kmp_table[candidate]
            candidate += 1
        kmp_table.append(candidate)

        limit = len(substring)+1 if must_be_suffix else len(substring)
        for i in range(limit):
            prefix_dict = transitions.setdefault(i, dict())
            for symbol in input_symbols:
                # Look for next state after reading in the given input symbol
                candidate = i if i < len(substring) else kmp_table[i]
                while candidate != -1 and substring[candidate] != symbol:
                    candidate = kmp_table[candidate]
                candidate += 1
                prefix_dict[symbol] = candidate

        states = frozenset(transitions.keys())
        final_states = {len(substring)}
        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states if contains else states - final_states,
        )

    @classmethod
    def from_subsequence(cls, input_symbols, subsequence, *, contains=True):
        """
        Directly computes the minimal DFA recognizing strings containing the
        given subsequence.
        If contains is set to False then the complement is constructed instead.
        """
        transitions = {0: {symbol: 0 for symbol in input_symbols}}

        for prev_state, char in enumerate(subsequence):
            next_state = prev_state + 1
            transitions[next_state] = {symbol: next_state for symbol in input_symbols}
            transitions[prev_state][char] = next_state

        states = frozenset(transitions.keys())
        final_states = {len(subsequence)}
        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states if contains else states - final_states,
        )

    @classmethod
    def of_length(cls, input_symbols, *, min_length=0, max_length=None, symbols_to_count=None):
        """
        Directly computes the minimal DFA recognizing strings whose length
        is between `min_length` and `max_length`, inclusive.
        To allow infinitely long words the value `None` can be passed in for `max_length`.
        """
        if symbols_to_count is None:
            symbols_to_count = input_symbols

        transitions = {}
        length_range = range(min_length) if max_length is None else range(max_length+1)
        for prev_state in length_range:
            next_state = prev_state + 1
            transitions[prev_state] = {
                symbol: next_state if symbol in symbols_to_count else prev_state
                for symbol in input_symbols
            }
        last_state = len(transitions)
        transitions[last_state] = {symbol: last_state for symbol in input_symbols}
        final_states = frozenset((last_state,)) if max_length is None else frozenset(range(min_length, max_length+1))

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states,
        )

    @classmethod
    def count_mod(cls, input_symbols, k, *, remainders=None, symbols_to_count=None):
        """
        Directly computes a DFA that counts given symbols and accepts all strings where
        the remainder of division by k is in the set of remainders given.
        The default value of remainders is {0} and all symbols are counted by default.
        """
        if k <= 0:
            raise ValueError("Integer must be positive")
        if symbols_to_count is None:
            symbols_to_count = input_symbols
        if remainders is None:
            remainders = {0}
        transitions = {i: {symbol: (i + 1) % k if symbol in symbols_to_count else i
                           for symbol in input_symbols}
                       for i in range(k)}

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=remainders
        )

    @classmethod
    def universal_language(cls, input_symbols):
        """
        Directly computes the minimal DFA accepting all strings.
        """
        return cls(
            states={0},
            input_symbols=input_symbols,
            transitions={0: {symbol: 0 for symbol in input_symbols}},
            initial_state=0,
            final_states={0}
        )

    @classmethod
    def empty_language(cls, input_symbols):
        """
        Directly computes the minimal DFA rejecting all strings.
        """
        return cls(
            states={0},
            input_symbols=input_symbols,
            transitions={0: {symbol: 0 for symbol in input_symbols}},
            initial_state=0,
            final_states=frozenset()
        )

    @classmethod
    def nth_from_start(cls, input_symbols, symbol, n):
        """
        Directly computes the minimal DFA which accepts all words whose `n`-th
        character from the start is `symbol`, where `n` is a positive integer.
        """
        if n < 1:
            raise ValueError("Integer must be positive")
        if symbol not in input_symbols:
            raise exceptions.InvalidSymbolError("Desired symbol is not in the set of input symbols")
        if len(input_symbols) == 1:
            return cls.of_length(input_symbols, min_length=n)
        transitions = {i: {symbol: i+1 for symbol in input_symbols} for i in range(n)}
        transitions[n-1][symbol] = n+1
        transitions[n] = {symbol: n for symbol in input_symbols}
        transitions[n+1] = {symbol: n+1 for symbol in input_symbols}

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states={n+1}
        )

    @classmethod
    def nth_from_end(cls, input_symbols, symbol, n):
        """
        Directly computes the minimal DFA which accepts all words whose `n`-th
        character from the end is `symbol`, where `n` is a positive integer.
        """
        if n < 1:
            raise ValueError("Integer must be positive")
        if symbol not in input_symbols:
            raise exceptions.InvalidSymbolError("Desired symbol is not in the set of input symbols")
        if len(input_symbols) == 1:
            return cls.of_length(input_symbols, min_length=n)
        # Consider the states to be labelled with bitstrings of size n
        # The bitstring represents how the current suffix in this state matches our desired symbol
        # A 1 means the character at this position is the desired symbol and a 0 means it is not
        # For transitions this is effectively doubling the label value and then adding 1 if the desired symbol is read
        # Finally we trim the label to n bits with a modulo operation.
        state_count = 2**n

        return cls(
            states=frozenset(range(state_count)),
            input_symbols=input_symbols,
            transitions={state: {sym: (2 * state + 1) % state_count
                                 if symbol == sym else (2 * state) % state_count
                                 for sym in input_symbols}
                         for state in range(state_count)},
            initial_state=0,
            final_states=frozenset(range(state_count//2, state_count)),
        )

    @classmethod
    def from_finite_language(cls, input_symbols, language):
        """
        Directly computes the minimal DFA corresponding to a finite language.
        Uses the algorithm described in Finite-State Techniques by Mihov and Schulz,
        Chapter 10
        """

        if not language:
            return DFA.empty_language(input_symbols)

        transitions = dict()
        back_map = {'': set()}
        final_states = set()
        signatures_dict = dict()

        def compute_signature(state):
            """Computes signature for input state"""
            return (state in final_states, frozenset(transitions[state].items()))

        def longest_common_prefix_length(string_1, string_2):
            """Returns length of longest common prefix."""
            for i, (chr_1, chr_2) in enumerate(zip(string_1, string_2)):
                if chr_1 != chr_2:
                    return i

            return min(len(string_1), len(string_2))

        def add_to_trie(word):
            """Add word to the trie represented by transitions"""
            prefix = ''
            for chr in word:
                next_prefix = prefix + chr

                # Extend the trie only if necessary
                prefix_dict = transitions.setdefault(prefix, dict())
                prefix_dict.setdefault(chr, next_prefix)
                back_map.setdefault(next_prefix, set()).add(prefix)

                prefix = next_prefix

            # Mark the finished prefix as a final state
            transitions[prefix] = dict()
            final_states.add(prefix)

        def compress(word, next_word):
            """Compress prefixes of word, newly added to the DFA"""

            # Compress along all prefixes that are _not_ prefixes of the next word
            lcp_len = longest_common_prefix_length(word, next_word)

            # Compression in order of reverse length
            for i in range(len(word), lcp_len, -1):
                prefix = word[:i]

                # Compute signature for comparison
                prefix_signature = compute_signature(prefix)
                identical_state = signatures_dict.get(prefix_signature)

                if identical_state is not None:
                    # If identical state exists, remove prefix since it's redundant
                    final_states.discard(prefix)
                    transitions.pop(prefix)

                    # Change transition for prefix
                    for parent_state in back_map[prefix]:
                        path = transitions[parent_state]
                        for chr in path:
                            if path[chr] == prefix:
                                path[chr] = identical_state
                        back_map[identical_state].add(parent_state)

                        # No need to recompute signatures here, since we will
                        # recompute when handling parents in later iterations
                else:
                    signatures_dict[prefix_signature] = prefix

        # Construct initial trie
        prev_word, *rest_words = sorted(language)
        add_to_trie(prev_word)

        for curr_word in rest_words:
            # For each word, compress from the previous iteration and continue
            compress(prev_word, curr_word)
            add_to_trie(curr_word)
            prev_word = curr_word

        compress(prev_word, '')

        # Add dump state. Always needed since dict is finite
        dump_state = 0
        transitions[dump_state] = {chr: dump_state for chr in input_symbols}

        for path in transitions.values():
            for chr in input_symbols:
                path.setdefault(chr, dump_state)

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state='',
            final_states=final_states,
        )

    @classmethod
    def from_nfa(cls, target_nfa, *, retain_names=False, minify=True):
        """Initialize this DFA as one equivalent to the given NFA."""
        # Data structures for state renaming
        new_state_name_dict = dict()
        state_name_counter = count(0)

        def get_name_renamed(states):
            nonlocal state_name_counter, new_state_name_dict
            return new_state_name_dict.setdefault(states, next(state_name_counter))

        def get_name_original(states):
            return states

        get_name = get_name_original if retain_names else get_name_renamed

        # equivalent DFA states states
        nfa_initial_states = frozenset(target_nfa._lambda_closures[target_nfa.initial_state])
        dfa_initial_state = get_name(nfa_initial_states)
        dfa_final_states = set()

        dfa_states = {dfa_initial_state}
        dfa_symbols = target_nfa.input_symbols
        dfa_transitions = {}

        state_queue = deque()
        state_queue.append(nfa_initial_states)
        while state_queue:
            current_states = state_queue.popleft()
            current_state_name = get_name(current_states)

            # Add NFA states to DFA as it is constructed from NFA.
            dfa_transitions[current_state_name] = {}
            if (current_states & target_nfa.final_states):
                dfa_final_states.add(current_state_name)

            # Enqueue the next set of current states for the generated DFA.
            for input_symbol in target_nfa.input_symbols:
                next_current_states = target_nfa._get_next_current_states(
                    current_states, input_symbol)

                next_current_states_name = get_name(next_current_states)
                dfa_transitions[current_state_name][input_symbol] = next_current_states_name

                # Only enqueue a state if it has not been seen yet.
                if next_current_states_name not in dfa_states:
                    dfa_states.add(next_current_states_name)
                    state_queue.append(next_current_states)

        if minify:
            return cls._minify(
                reachable_states=dfa_states,
                input_symbols=dfa_symbols,
                transitions=dfa_transitions,
                initial_state=dfa_initial_state,
                reachable_final_states=dfa_final_states,
                retain_names=retain_names)

        return cls(
            states=dfa_states,
            input_symbols=dfa_symbols,
            transitions=dfa_transitions,
            initial_state=dfa_initial_state,
            final_states=dfa_final_states)

    def show_diagram(self, path=None):
        """
            Creates the graph associated with this DFA
        """
        # Nodes are set of states

        graph = Dot(graph_type='digraph', rankdir='LR')
        nodes = {}
        for state in self.states:
            if state == self.initial_state:
                # color start state with green
                if state in self.final_states:
                    initial_state_node = Node(
                        state,
                        style='filled',
                        peripheries=2,
                        fillcolor='#66cc33')
                else:
                    initial_state_node = Node(
                        state, style='filled', fillcolor='#66cc33')
                nodes[state] = initial_state_node
                graph.add_node(initial_state_node)
            else:
                if state in self.final_states:
                    state_node = Node(state, peripheries=2)
                else:
                    state_node = Node(state)
                nodes[state] = state_node
                graph.add_node(state_node)
        # adding edges
        for from_state, lookup in self.transitions.items():
            for to_label, to_state in lookup.items():
                graph.add_edge(Edge(
                    nodes[from_state],
                    nodes[to_state],
                    label=to_label
                ))
        if path:
            graph.write_png(path)
        return graph



#!/usr/bin/env python3
"""Miscellaneous utility functions and classes."""

from frozendict import frozendict


def freezeValue(value):
    """
    A helper function to convert the given value / structure into a fully
    immutable one by recursively processing said structure and any of its
    members, freezing them as well
    """
    if isinstance(value, list):
        return tuple(freezeValue(element) for element in value)
    if isinstance(value, set):
        return frozenset(freezeValue(element) for element in value)
    elif isinstance(value, dict):
        return frozendict({
            dict_key: freezeValue(dict_value)
            for dict_key, dict_value in value.items()
        })
    else:
        return value


class PartitionRefinement:
    """Maintain and refine a partition of a set of items into subsets.
    Space usage for a partition of n items is O(n), and each refine
    operation takes time proportional to the size of its argument.

    Adapted from code by D. Eppstein: https://www.ics.uci.edu/~eppstein/PADS/PartitionRefinement.py
    """

    __slots__ = ('_sets', '_partition')

    def __init__(self, items):
        """Create a new partition refinement data structure for the given
        items. Initially, all items belong to the same subset.
        """
        S = set(items)
        self._sets = {id(S): S}
        self._partition = {x: id(S) for x in S}

    def get_set_by_id(self, id):
        """Return the set in the partition corresponding to id."""
        return self._sets[id]

    def get_set_ids(self):
        """Return set ids corresponding to the internal partition."""
        return self._sets.keys()

    def get_sets(self):
        """Return sets corresponding to the internal partition."""
        return self._sets.values()

    def refine(self, S):
        """Refine each set A in the partition to the two sets
        A & S, A - S.  Return a list of pairs ids (id(A & S), id(A - S))
        for each changed set.  Within each pair, A & S will be
        a newly created set, while A - S will be a modified
        version of an existing set in the partition (retaining its old id).
        Not a generator because we need to perform the partition
        even if the caller doesn't iterate through the results.
        """
        hit = {}
        output = []

        for x in S:
            Aid = self._partition[x]
            hit.setdefault(Aid, set()).add(x)

        for Aid, AS in hit.items():
            A = self._sets[Aid]
            if AS != A:
                self._sets[id(AS)] = AS
                for x in AS:
                    self._partition[x] = id(AS)
                A -= AS
                output.append((id(AS), Aid))

        return output



#!/usr/bin/env python3
"""Classes and methods for working with generalized non-deterministic finite automata."""

from itertools import product

from frozendict import frozendict
from pydot import Dot, Edge, Node

import automata.base.exceptions as exceptions
import automata.fa.fa as fa
import automata.fa.nfa as nfa
import automata.regex.regex as re


class GNFA(nfa.NFA):
    """A generalized nondeterministic finite automaton."""

    # The conventions of using __slots__ state that subclasses automatically
    # inherit __slots__ from parent classes, so there's no need to redeclare
    # slotted attributes for each subclass; however, because NFA has a
    # 'final_states' attribute but GNFA has a 'final_state' attribute, we must
    # redeclare them below to exclude 'final_states'
    __slots__ = ('states', 'input_symbols', 'transitions',
                 'initial_state', 'final_state')

    def __init__(self, *, states, input_symbols, transitions,
                 initial_state, final_state):
        """Initialize a complete NFA."""
        super(fa.FA, self).__init__(
            states=frozenset(states),
            input_symbols=frozenset(input_symbols),
            transitions=frozendict({
                state: frozendict(paths)
                for state, paths in transitions.items()
            }),
            initial_state=initial_state,
            final_state=final_state
        )

    # GNFA should NOT create the lambda closures via NFA.__post_init__()
    def __post_init__(self):
        self.validate()

    @classmethod
    def from_dfa(cls, dfa):
        """Initialize this GNFA as one equivalent to the given DFA."""
        new_gnfa_transitions = dict()
        gnfa_states = set(dfa.states)

        for state in dfa.states:
            gnfa_transitions = dict()
            if state in dfa.transitions:
                for input_symbol, to_state in dfa.transitions[state].items():
                    if to_state in gnfa_transitions.keys():
                        gnfa_transitions[to_state] = f"{gnfa_transitions[to_state]}|{input_symbol}"
                    else:
                        gnfa_transitions[to_state] = input_symbol
                new_gnfa_transitions[state] = gnfa_transitions
            else:
                new_gnfa_transitions[state] = dict()

        new_initial_state = GNFA._add_new_state(gnfa_states)
        new_final_state = GNFA._add_new_state(gnfa_states, new_initial_state)

        new_gnfa_transitions[new_initial_state] = {dfa.initial_state: ''}

        for state in dfa.final_states:
            new_gnfa_transitions[state][new_final_state] = ''

        for state in gnfa_states - {new_final_state}:  # pragma: no branch
            if gnfa_states - new_gnfa_transitions[state].keys():  # pragma: no branch
                for leftover_state in gnfa_states - new_gnfa_transitions[state].keys():
                    if leftover_state is not new_initial_state:
                        new_gnfa_transitions[state][leftover_state] = None

        return cls(
            states=gnfa_states, input_symbols=dfa.input_symbols,
            transitions=new_gnfa_transitions, initial_state=new_initial_state,
            final_state=new_final_state)

    @classmethod
    def from_nfa(cls, nfa):
        """Initialize this GNFA as one equivalent to the given NFA."""
        new_gnfa_transitions = dict()
        gnfa_states = set(nfa.states)

        for state in nfa.states:
            gnfa_transitions = dict()
            if state in nfa.transitions:
                for input_symbol, to_states in nfa.transitions[state].items():
                    for to_state in to_states:
                        if to_state in gnfa_transitions.keys():
                            if gnfa_transitions[to_state] == '' and input_symbol != '':
                                gnfa_transitions[to_state] = f'{input_symbol}?'
                            elif gnfa_transitions[to_state] != '' and input_symbol == '':
                                if cls._isbracket_req(gnfa_transitions[to_state]):
                                    gnfa_transitions[to_state] = f'({gnfa_transitions[to_state]})?'
                                else:
                                    gnfa_transitions[to_state] = f'{gnfa_transitions[to_state]}?'
                            else:
                                gnfa_transitions[to_state] = f"{gnfa_transitions[to_state]}|{input_symbol}"
                        else:
                            gnfa_transitions[to_state] = input_symbol
                new_gnfa_transitions[state] = gnfa_transitions
            else:
                new_gnfa_transitions[state] = dict()

        new_initial_state = GNFA._add_new_state(gnfa_states)
        new_final_state = GNFA._add_new_state(gnfa_states, new_initial_state)

        new_gnfa_transitions[new_initial_state] = {nfa.initial_state: ''}

        for state in nfa.final_states:
            new_gnfa_transitions[state][new_final_state] = ''

        for state in gnfa_states - {new_final_state}:  # pragma: no branch
            if gnfa_states - new_gnfa_transitions[state].keys():  # pragma: no branch
                for leftover_state in gnfa_states - new_gnfa_transitions[state].keys():
                    if leftover_state is not new_initial_state:
                        new_gnfa_transitions[state][leftover_state] = None

        return cls(
            states=gnfa_states, input_symbols=nfa.input_symbols,
            transitions=new_gnfa_transitions, initial_state=new_initial_state,
            final_state=new_final_state)

    def _validate_transition_invalid_symbols(self, start_state, paths):
        """Raise an error if transition symbols are invalid."""
        for regex in paths.values():
            check = self.input_symbols.copy()
            check = check.union({'*', '|', '(', ')', '?'})
            if regex is not None and (set(regex) - check and regex != '' or not re._validate(regex)):
                raise exceptions.InvalidRegexError(
                    'state {} has invalid transition expression {}'.format(
                        start_state, regex))

    def _validate_transition_end_states(self, start_state, paths):
        """Raise an error if transition end states are invalid or missing"""
        if start_state == self.final_state:
            if len(paths) != 0:  # pragma: no branch
                raise exceptions.InvalidStateError(
                    'No transitions should be defined for '
                    'final state {}'.format(start_state))
        elif start_state == self.initial_state and self.states - paths.keys() - {self.initial_state} != set():
            raise exceptions.MissingStateError(
                'state {} does not have transitions defined for states {}'.format(
                    start_state, str(self.states - paths.keys() - {self.initial_state})))
        elif start_state != self.initial_state and self.states - paths.keys() - {self.initial_state} != set():
            raise exceptions.MissingStateError(
                'state {} does not have transitions defined for states {}'.format(
                    start_state, str(self.states - paths.keys() - {self.initial_state})))
        for end_state in paths.keys():  # pragma: no branch
            if end_state not in self.states:
                raise exceptions.InvalidStateError(
                    'end state {} for transition on {} is '
                    'not valid'.format(end_state, start_state))

    def _validate_final_state(self):
        """Raise an error if the initial state is invalid."""
        if self.final_state not in self.states:
            raise exceptions.InvalidStateError(
                '{} is not a valid final state'.format(self.final_state))

    def validate(self):
        """Return True if this NFA is internally consistent."""
        self._validate_initial_state()
        self._validate_final_state()
        for start_state, paths in self.transitions.items():
            self._validate_transition_invalid_symbols(start_state, paths)
            self._validate_transition_end_states(start_state, paths)
        self._validate_initial_state_transitions()
        return True

    @staticmethod
    def _isbracket_req(regex):
        bracket_open = 0
        for i in range(len(regex)):
            if regex[i] == '(':
                bracket_open += 1
            elif regex[i] == ')':
                bracket_open = bracket_open - 1
            if bracket_open == 0 and regex[i] == '|':
                return True

        return False

    @staticmethod
    def _find_min_connected_node(states, transitions, initial_state, final_state):
        state_set = states-{initial_state, final_state}
        state_degree = dict.fromkeys(state_set, 0)
        for state in states - {final_state}:
            for to_state, label in transitions[state].items():
                if label is not None:
                    if state != initial_state:
                        state_degree[state] = state_degree[state] + 1
                    if to_state != final_state:
                        state_degree[to_state] = state_degree[to_state] + 1

        return min(state_degree, key=state_degree.get)

    def to_regex(self):
        """
        Convert GNFA to regular expression.
        """
        new_states = set(self.states)
        new_transitions = {
            state: dict(paths)
            for state, paths in self.transitions.items()
        }

        while len(new_states) > 2:
            q_rip = self._find_min_connected_node(new_states, new_transitions, self.initial_state, self.final_state)
            new_states.remove(q_rip)
            for q_i, q_j in product(new_states - {self.final_state}, new_states - {self.initial_state}):
                r1 = new_transitions[q_i][q_rip]
                r2 = new_transitions[q_rip][q_rip]
                r3 = new_transitions[q_rip][q_j]
                r4 = new_transitions[q_i][q_j]

                if r1 is None or r3 is None:
                    new_transitions[q_i][q_j] = r4
                else:
                    # check if putting brackets around r1 is necessary
                    if self._isbracket_req(r1):
                        r1 = f'({r1})'

                    if r2 is None:
                        r2 = ''
                    elif len(r2) == 1:
                        r2 = f'{r1}*'
                    else:
                        r2 = f'({r2})*'

                    if self._isbracket_req(r3):
                        r3 = f'({r3})'

                    if r4 is None:
                        r4 = ''
                    elif self._isbracket_req(r4):
                        r4 = f'|({r4})'
                    elif r4 == '':
                        r4 = '?'
                    else:
                        r4 = f'|{r4}'

                    if r4 == '?' and len(r1+r2+r3) > 1:
                        new_transitions[q_i][q_j] = f'({r1}{r2}{r3}){r4}'
                    else:
                        new_transitions[q_i][q_j] = f'{r1}{r2}{r3}{r4}'

            del new_transitions[q_rip]
            for state in new_states-{self.final_state}:
                del new_transitions[state][q_rip]

        return new_transitions[self.initial_state][self.final_state]

    def __eq__(self, other):
        return NotImplemented

    # The following NFA methods are not supported on GNFA instances because
    # they are out of scope for the purpose of the GNFA class (which is focused
    # on regular expression conversion)
    def read_input_stepwise(self, input_str): raise NotImplementedError
    def union(self, other): raise NotImplementedError
    def concatenate(self, other): raise NotImplementedError
    def kleene_star(self): raise NotImplementedError
    def option(self): raise NotImplementedError
    def reverse(self): raise NotImplementedError

    def show_diagram(self, path=None, show_None=True):
        """
            Creates the graph associated with this DFA
        """
        # Nodes are set of states

        graph = Dot(graph_type='digraph', rankdir='LR')
        nodes = {}
        for state in self.states:
            if state == self.initial_state:
                # color start state with green
                initial_state_node = Node(
                    state, style='filled', fillcolor='#66cc33')
                nodes[state] = initial_state_node
                graph.add_node(initial_state_node)
            else:
                if state == self.final_state:
                    state_node = Node(state, peripheries=2)
                else:
                    state_node = Node(state)
                nodes[state] = state_node
                graph.add_node(state_node)
        # adding edges
        for from_state, lookup in self.transitions.items():
            for to_state, to_label in lookup.items():  # pragma: no branch
                if to_label is None and show_None:
                    to_label = "ø"
                    graph.add_edge(Edge(
                        nodes[from_state],
                        nodes[to_state],
                        label=to_label
                    ))
                elif to_label is not None:
                    graph.add_edge(Edge(
                        nodes[from_state],
                        nodes[to_state],
                        label=to_label
                    ))
        if path:
            graph.write_png(path)
        return graph



#!/usr/bin/env python3
"""
Classes and methods for lexing expressions into lists of tokens.

Adapted from code in the tdparser library: https://github.com/rbarrois/tdparser
"""

import abc
import re

import automata.base.exceptions as exceptions


class Token(metaclass=abc.ABCMeta):
    """Base class for tokens."""

    __slots__ = ('text',)

    def __init__(self, text):
        self.text = text

    def get_precedence(self):
        raise NotImplementedError

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.text}>"


class TokenRegistry():
    """Registry holding token rules."""

    __slots__ = ('_tokens',)

    def __init__(self):
        self._tokens = []

    def register(self, token_factory_fn, token_regex):
        """
        Register a token that can be produced by token_factory_fn (a function
        taking in a string returning the final token) and recognized by the
        token_regex pattern.
        """
        self._tokens.append((token_factory_fn, re.compile(token_regex)))

    def matching_tokens(self, text, start):
        """Retrieve all token definitions matching text starting at start."""

        for token_factory_fn, regexp in self._tokens:
            match = regexp.match(text, pos=start)
            if match:
                yield (token_factory_fn, match)

    def get_token(self, text, start=0):
        """
        Retrieve the next token from some text. Computes the best match by
        length. Returns None if there is no match in the token registry.
        """
        best_token_match = None
        best_match = None

        for token_factory_fn, match in self.matching_tokens(text, start):
            if not best_match or best_match.end() < match.end():
                best_token_match = (token_factory_fn, match)
                best_match = match

        return best_token_match

    def __len__(self):
        return len(self._tokens)


class Lexer():
    """
    The core lexer. First, tokens are registered with their factory functions and regex
    patterns. The lexer can then take in a string and splits it into a list of token
    classes (in infix ordering) matching the regex patterns.
    """

    __slots__ = ('tokens', 'blank_chars')

    def __init__(self, blank_chars={' ', '\t'}):
        self.tokens = TokenRegistry()
        self.blank_chars = blank_chars

    def register_token(self, token_factory_fn, token_regex):
        """
        Register a token class. The token_factory_fn must taken in a
        string and return an instance of the desired token, and token_regex
        is used by the lexer to match tokens in the input text.
        """

        self.tokens.register(token_factory_fn, token_regex)

    def lex(self, text):
        """Split text into a list of tokens in infix notation."""

        pos = 0
        res = []

        while pos < len(text):
            token_match = self.tokens.get_token(text, start=pos)
            if token_match is not None:
                token_factory_fn, match = token_match
                matched_text = match.group(0)
                res.append(token_factory_fn(matched_text))
                pos += len(matched_text)
            elif text[pos] in self.blank_chars:
                pos += 1
            else:
                raise exceptions.LexerError(f"Invalid character '{text[pos]}' in '{text}'", position=pos)

        return res



#!/usr/bin/env python3
"""Methods for working with regular expressions"""

import automata.base.exceptions as exceptions
from automata.fa.nfa import NFA
from automata.regex.parser import (RESERVED_CHARACTERS, get_regex_lexer,
                                   validate_tokens)


def _validate(regex):
    """Return True if the regular expression is valid"""

    try:
        validate(regex)
    except exceptions.InvalidRegexError:
        return False

    return True


def validate(regex):
    """Raise an error if the regular expression is invalid"""
    input_symbols = set(regex) - RESERVED_CHARACTERS

    validate_tokens(get_regex_lexer(input_symbols).lex(regex))

    return True


def isequal(re1, re2, *, input_symbols=None):
    """Return True if both regular expressions are equivalent"""

    nfa1 = NFA.from_regex(re1, input_symbols=input_symbols)
    nfa2 = NFA.from_regex(re2, input_symbols=input_symbols)

    return nfa1 == nfa2


def issubset(re1, re2, *, input_symbols=None):
    """Return True if re1 is a subset of re2"""

    nfa1 = NFA.from_regex(re1, input_symbols=input_symbols)
    nfa2 = NFA.from_regex(re2, input_symbols=input_symbols)

    return nfa1.union(nfa2) == nfa2


def issuperset(re1, re2, *, input_symbols=None):
    """Return True if re1 is a subset of re2"""

    nfa1 = NFA.from_regex(re1, input_symbols=input_symbols)
    nfa2 = NFA.from_regex(re2, input_symbols=input_symbols)

    return nfa1.union(nfa2) == nfa1



#!/usr/bin/env python3
"""Classes and methods for parsing regexes into NFAs."""

from collections import deque
from itertools import chain, count, product, zip_longest

from automata.regex.lexer import Lexer
from automata.regex.postfix import (InfixOperator, LeftParen, Literal,
                                    PostfixOperator, RightParen,
                                    parse_postfix_tokens, tokens_to_postfix,
                                    validate_tokens)

RESERVED_CHARACTERS = frozenset({'*', '|', '(', ')', '?', ' ', '\t', '&', '+', '.'})


class NFARegexBuilder:
    """Builder class designed for speed in parsing regular expressions into NFAs."""

    __slots__ = ('_transitions', '_initial_state', '_final_states')
    _state_name_counter = count(0)

    def __init__(self, *, transitions, initial_state, final_states):
        """
        Initialize new builder class
        """

        self._transitions = transitions
        self._initial_state = initial_state
        self._final_states = final_states

    @classmethod
    def from_string_literal(cls, literal):
        """
        Initialize this builder accepting only the given string literal
        """

        transitions = {
            cls.__get_next_state_name(): {chr: set()}
            for chr in literal
        }

        for start_state, path in transitions.items():
            for end_states in path.values():
                end_states.add(start_state+1)

        final_state = cls.__get_next_state_name()
        transitions[final_state] = dict()

        return cls(
            transitions=transitions,
            initial_state=min(transitions.keys()),
            final_states={final_state}
        )

    @classmethod
    def wildcard(cls, input_symbols):
        """
        Initialize this builder for a wildcard with the given input symbols
        """

        initial_state = cls.__get_next_state_name()
        final_state = cls.__get_next_state_name()

        transitions = {
            initial_state: {symbol: {final_state} for symbol in input_symbols},
            final_state: dict()
        }

        return cls(
            transitions=transitions,
            initial_state=initial_state,
            final_states={final_state}
        )

    def union(self, other):
        """
        Apply the union operation to the NFA represented by this builder and other
        """
        self._transitions.update(other._transitions)

        new_initial_state = self.__get_next_state_name()

        # Add epsilon transitions from new start state to old ones
        self._transitions[new_initial_state] = {
            '': {self._initial_state, other._initial_state}
        }

        self._initial_state = new_initial_state
        self._final_states.update(other._final_states)

    def intersection(self, other):
        """
        Apply the intersection operation to the NFA represented by this builder and other.
        Use BFS to only traverse reachable part (keeps number of states down).
        """
        new_state_name_dict = dict()

        def get_state_name(state_name):
            return new_state_name_dict.setdefault(state_name, self.__get_next_state_name())

        new_final_states = set()
        new_transitions = dict()
        new_initial_state = (self._initial_state, other._initial_state)

        new_initial_state_name = get_state_name(new_initial_state)
        new_input_symbols = set(chain.from_iterable(
            transition_dict.keys()
            for transition_dict in chain(self._transitions.values(), other._transitions.values())
        )) - {''}

        queue = deque()

        queue.append(new_initial_state)
        new_transitions[new_initial_state_name] = dict()

        while queue:
            curr_state = queue.popleft()
            curr_state_name = get_state_name(curr_state)
            q_a, q_b = curr_state

            if q_a in self._final_states and q_b in other._final_states:
                new_final_states.add(curr_state_name)

            # States we will consider adding to the queue
            next_states_iterables = list()

            # Get transition dict for states in self
            transitions_a = self._transitions.get(q_a, {})
            # Add epsilon transitions for first set of transitions
            epsilon_transitions_a = transitions_a.get('')
            if epsilon_transitions_a is not None:
                state_dict = new_transitions.setdefault(curr_state_name, dict())
                state_dict.setdefault('', set()).update(
                    get_state_name(state) for state in product(epsilon_transitions_a, [q_b])
                )
                next_states_iterables.append(product(epsilon_transitions_a, [q_b]))

            # Get transition dict for states in other
            transitions_b = other._transitions.get(q_b, {})
            # Add epsilon transitions for second set of transitions
            epsilon_transitions_b = transitions_b.get('')
            if epsilon_transitions_b is not None:
                state_dict = new_transitions.setdefault(curr_state_name, dict())
                state_dict.setdefault('', set()).update(
                    get_state_name(state) for state in product([q_a], epsilon_transitions_b)
                )
                next_states_iterables.append(product([q_a], epsilon_transitions_b))

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state_name, dict())
                    state_dict.setdefault(symbol, set()).update(
                        get_state_name(state) for state in product(end_states_a, end_states_b)
                    )
                    next_states_iterables.append(product(end_states_a, end_states_b))

            # Finally, try visiting every state we found.
            for product_state in chain.from_iterable(next_states_iterables):
                product_state_name = get_state_name(product_state)
                if product_state_name not in new_transitions:
                    new_transitions[product_state_name] = dict()
                    queue.append(product_state)

        self._final_states = new_final_states
        self._transitions = new_transitions
        self._initial_state = new_initial_state_name

    def concatenate(self, other):
        """
        Apply the concatenate operation to the NFA represented by this builder
        and other.
        """
        self._transitions.update(other._transitions)

        for state in self._final_states:
            self._transitions[state].setdefault('', set()).add(other._initial_state)

        self._final_states = other._final_states

    def kleene_star(self):
        """
        Apply the kleene star operation to the NFA represented by this builder
        """
        new_initial_state = self.__get_next_state_name()

        self._transitions[new_initial_state] = {
            '': {self._initial_state}
        }

        for state in self._final_states:
            self._transitions[state].setdefault('', set()).add(self._initial_state)

        self._initial_state = new_initial_state
        self._final_states.add(new_initial_state)

    def kleene_plus(self):
        """
        Apply the kleene plus operation to the NFA represented by this builder
        """
        state_map = {
            old_state_name: self.__get_next_state_name()
            for old_state_name in self._transitions
        }

        second_initial_state = state_map[self._initial_state]

        new_transitions = {
            state_map[state]: {
                chr: {state_map[end_state] for end_state in end_states}
                for chr, end_states in path.items()
            }
            for state, path in self._transitions.items()
        }

        new_initial_state = self.__get_next_state_name()

        new_transitions[new_initial_state] = {
            '': {second_initial_state}
        }

        new_final_states = {new_initial_state}
        for state in self._final_states:
            # Add transitions from old final states to intermediate one
            self._transitions[state].setdefault('', set()).add(new_initial_state)

            # Add transitions from the new final states to new initial state
            new_final_state = state_map[state]
            new_final_states.add(new_final_state)
            new_transitions[new_final_state].setdefault('', set()).add(new_initial_state)

        self._transitions.update(new_transitions)
        self._final_states = new_final_states

    def option(self):
        """
        Apply the option operation to the NFA represented by this builder
        """
        new_initial_state = self.__get_next_state_name()

        self._transitions[new_initial_state] = {
            '': {self._initial_state}
        }

        self._initial_state = new_initial_state
        self._final_states.add(new_initial_state)

    @classmethod
    def __get_next_state_name(cls):
        return next(cls._state_name_counter)


class UnionToken(InfixOperator):
    """Subclass of infix operator defining the union operator."""

    def get_precedence(self):
        return 1

    def op(self, left, right):
        left.union(right)
        return left


class IntersectionToken(InfixOperator):
    """Subclass of infix operator defining the intersection operator."""

    def get_precedence(self):
        return 1

    def op(self, left, right):
        left.intersection(right)
        return left


class KleeneStarToken(PostfixOperator):
    """Subclass of postfix operator defining the kleene star operator."""

    def get_precedence(self):
        return 3

    def op(self, left):
        left.kleene_star()
        return left


class KleenePlusToken(PostfixOperator):
    """Subclass of postfix operator defining the kleene plus operator."""

    def get_precedence(self):
        return 3

    def op(self, left):
        left.kleene_plus()
        return left


class OptionToken(PostfixOperator):
    """Subclass of postfix operator defining the option operator."""

    def get_precedence(self):
        return 3

    def op(self, left):
        left.option()
        return left


class ConcatToken(InfixOperator):
    """Subclass of infix operator defining the concatenation operator."""

    def get_precedence(self):
        return 2

    def op(self, left, right):
        left.concatenate(right)
        return left


class StringToken(Literal):
    """Subclass of literal token defining a string literal."""

    def val(self):
        return NFARegexBuilder.from_string_literal(self.text)


class WildcardToken(Literal):
    """Subclass of literal token defining a wildcard literal."""

    def __init__(self, text, input_symbols):
        super().__init__(text)
        self.input_symbols = input_symbols

    def val(self):
        return NFARegexBuilder.wildcard(self.input_symbols)


def add_concat_tokens(token_list):
    """Add concat tokens to list of parsed infix tokens."""

    final_token_list = []

    # Pairs of token types to insert concat tokens in between
    concat_pairs = [
        (Literal, Literal),
        (RightParen, LeftParen),
        (RightParen, Literal),
        (Literal, LeftParen),
        (PostfixOperator, Literal),
        (PostfixOperator, LeftParen)
    ]

    for curr_token, next_token in zip_longest(token_list, token_list[1:]):
        final_token_list.append(curr_token)

        if next_token:
            for firstClass, secondClass in concat_pairs:
                if isinstance(curr_token, firstClass) and isinstance(next_token, secondClass):
                    final_token_list.append(ConcatToken(''))

    return final_token_list


def get_regex_lexer(input_symbols):
    """Get lexer for parsing regular expressions."""
    lexer = Lexer()

    lexer.register_token(LeftParen, r'\(')
    lexer.register_token(RightParen, r'\)')
    lexer.register_token(StringToken, r'[A-Za-z0-9]')
    lexer.register_token(UnionToken, r'\|')
    lexer.register_token(IntersectionToken, r'\&')
    lexer.register_token(KleeneStarToken, r'\*')
    lexer.register_token(KleenePlusToken, r'\+')
    lexer.register_token(OptionToken, r'\?')
    lexer.register_token(lambda text: WildcardToken(text, input_symbols), r'\.')

    return lexer


def parse_regex(regexstr, input_symbols):
    """Return an NFARegexBuilder corresponding to regexstr."""

    if len(regexstr) == 0:
        return NFARegexBuilder.from_string_literal(regexstr)

    lexer = get_regex_lexer(input_symbols)
    lexed_tokens = lexer.lex(regexstr)
    validate_tokens(lexed_tokens)
    tokens_with_concats = add_concat_tokens(lexed_tokens)
    postfix = tokens_to_postfix(tokens_with_concats)

    return parse_postfix_tokens(postfix)



#!/usr/bin/env python3
"""Classes and methods for converting lists of tokens to postfix ordering."""

import abc
from collections import deque
from itertools import zip_longest

import automata.base.exceptions as exceptions
from automata.regex.lexer import Token


class Operator(Token):
    """Subclass of token defining an operator."""

    @abc.abstractmethod
    def get_precedence(self):
        raise NotImplementedError


class InfixOperator(Operator):
    """Subclass of operator defining an infix operator."""

    @abc.abstractmethod
    def op(self, left, right):
        raise NotImplementedError


class PostfixOperator(Operator):
    """Subclass of operator defining an postfix operator."""

    @abc.abstractmethod
    def op(self, left):
        raise NotImplementedError


class Literal(Token):
    """Subclass of token defining a literal."""

    @abc.abstractmethod
    def val(self):
        raise NotImplementedError


class RightParen(Token):
    """Subclass of token defining a right parenthesis."""

    def __repr__(self):
        return '<)>'


class LeftParen(Token):
    """Subclass of token defining a left parenthesis."""

    def __repr__(self):
        return '<(>'


def validate_tokens(token_list):
    """Validate the inputted tokens list (in infix ordering)."""

    token_list_prev = [None]
    token_list_prev.extend(token_list)

    paren_counter = 0

    for prev_token, curr_token in zip_longest(token_list_prev, token_list):
        # No postfix or infix operators at the beginning
        if prev_token is None and isinstance(curr_token, (InfixOperator, PostfixOperator)):
            raise exceptions.InvalidRegexError(f"Token '{curr_token}' cannot appear at the start of a statement.")

        # No postfix operators at the end of a statement or right before another operator or right paren
        elif isinstance(prev_token, InfixOperator):
            if curr_token is None:
                raise exceptions.InvalidRegexError(f"'{prev_token}' cannot appear at the end of a statement.")
            elif isinstance(curr_token, (InfixOperator, PostfixOperator, RightParen)):
                raise exceptions.InvalidRegexError(f"'{prev_token}' cannot appear immediately before '{curr_token}'.")

        # No left parens right before infix or postfix operators, or right before a right paren
        elif isinstance(prev_token, LeftParen):
            if isinstance(curr_token, (InfixOperator, PostfixOperator, RightParen)):
                raise exceptions.InvalidRegexError(f"'{prev_token}' cannot appear immediately before '{prev_token}'.")

            # Track open/closed parens
            paren_counter += 1

        elif isinstance(prev_token, RightParen):
            paren_counter -= 1

            if paren_counter < 0:
                raise exceptions.InvalidRegexError("Token list has mismatched parethesis.")

    if paren_counter != 0:
        raise exceptions.InvalidRegexError("Token list has unclosed parethesis.")


def tokens_to_postfix(tokens):
    """Takes in a list of tokens and changes them to postfix ordering."""
    stack = deque()
    res = []

    def comp_precedence(a, b):
        """Compare precedence of operators (two tokens)."""
        return a.get_precedence() <= b.get_precedence()

    for c in tokens:
        if isinstance(c, Literal):
            res.append(c)
        elif isinstance(c, RightParen):
            while len(stack) > 0 and not isinstance(stack[-1], LeftParen):
                res.append(stack.pop())
            stack.pop()
        elif isinstance(c, LeftParen):
            stack.append(c)
        elif not stack or isinstance(stack[-1], LeftParen) or not comp_precedence(c, stack[-1]):
            stack.append(c)
        else:
            while stack and not isinstance(stack[-1], LeftParen) and comp_precedence(c, stack[-1]):
                res.append(stack.pop())
            stack.append(c)

    while stack:
        res.append(stack.pop())

    return res


def parse_postfix_tokens(postfix_tokens):
    """Parse list of postfix tokens to produce value of expression."""

    stack = deque()

    for token in postfix_tokens:
        if isinstance(token, InfixOperator):
            right = stack.pop()
            left = stack.pop()
            stack.append(token.op(left, right))
        elif isinstance(token, PostfixOperator):
            left = stack.pop()
            stack.append(token.op(left))
        elif isinstance(token, Literal):
            stack.append(token.val())
        else:
            raise exceptions.InvalidRegexError(f"Invalid token type {type(token)}")

    return stack[0]
